# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import math
import string
import contextlib
import functools
import logging
from abc import ABC, abstractmethod
from typing import List, Tuple, Optional, Union, Dict, Generator, Iterator

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3 import framework
from volatility3.framework import (
    constants,
    exceptions,
    deprecation,
    interfaces,
    objects,
)
from volatility3.framework.objects import utility
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.linux import extensions
from volatility3.framework.layers import scanners
from volatility3.framework.constants import linux as linux_constants

vollog = logging.getLogger(__name__)


class LinuxKernelIntermedSymbols(intermed.IntermediateSymbolTable):
    provides = {"type": "interface"}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Set-up Linux specific types
        self.set_type_class("file", extensions.struct_file)
        self.set_type_class("list_head", extensions.list_head)
        self.set_type_class("hlist_head", extensions.hlist_head)
        self.set_type_class("mm_struct", extensions.mm_struct)
        self.set_type_class("super_block", extensions.super_block)
        self.set_type_class("task_struct", extensions.task_struct)
        self.set_type_class("vm_area_struct", extensions.vm_area_struct)
        self.set_type_class("qstr", extensions.qstr)
        self.set_type_class("dentry", extensions.dentry)
        self.set_type_class("fs_struct", extensions.fs_struct)
        self.set_type_class("files_struct", extensions.files_struct)
        self.set_type_class("kobject", extensions.kobject)
        self.set_type_class("cred", extensions.cred)
        self.set_type_class("inode", extensions.inode)
        self.set_type_class("idr", extensions.IDR)
        self.set_type_class("address_space", extensions.address_space)
        self.set_type_class("page", extensions.page)

        # Might not exist in the current symbols
        self.optional_set_type_class("module", extensions.module)
        self.optional_set_type_class("bpf_prog", extensions.bpf_prog)
        self.optional_set_type_class("bpf_prog_aux", extensions.bpf_prog_aux)
        self.optional_set_type_class("kernel_cap_struct", extensions.kernel_cap_struct)
        self.optional_set_type_class("kernel_cap_t", extensions.kernel_cap_t)
        self.optional_set_type_class("scatterlist", extensions.scatterlist)
        self.optional_set_type_class("module_sect_attr", extensions.module_sect_attr)
        self.optional_set_type_class("bin_attribute", extensions.bin_attribute)

        # kernels >= 4.18
        self.optional_set_type_class("timespec64", extensions.timespec64)
        # kernels < 4.18. Reuses timespec64 obj extension, since both has the same members
        self.optional_set_type_class("timespec", extensions.timespec64)

        # Mount
        self.set_type_class("vfsmount", extensions.vfsmount)
        # Might not exist in older kernels or the current symbols
        self.optional_set_type_class("mount", extensions.mount)
        self.optional_set_type_class("mnt_namespace", extensions.mnt_namespace)
        self.optional_set_type_class("rb_root", extensions.rb_root)

        # Network
        # FIXME: Deprecate all of this once the framework hits version 3
        self.set_type_class("net", extensions.network.net)
        self.set_type_class("socket", extensions.network.socket)
        self.set_type_class("sock", extensions.network.sock)
        self.set_type_class("inet_sock", extensions.network.inet_sock)
        self.set_type_class("unix_sock", extensions.network.unix_sock)

        # Might not exist in older kernels or the current symbols
        self.optional_set_type_class("netlink_sock", extensions.network.netlink_sock)
        self.optional_set_type_class("vsock_sock", extensions.network.vsock_sock)
        self.optional_set_type_class("packet_sock", extensions.network.packet_sock)
        self.optional_set_type_class("bt_sock", extensions.network.bt_sock)
        self.optional_set_type_class("xdp_sock", extensions.network.xdp_sock)

        # Only found in 6.1+ kernels
        self.optional_set_type_class("maple_tree", extensions.maple_tree)

        self.optional_set_type_class("latch_tree_root", extensions.latch_tree_root)
        self.optional_set_type_class("kernel_symbol", extensions.kernel_symbol)


class LinuxUtilities(interfaces.configuration.VersionableInterface):
    """Class with multiple useful linux functions."""

    _version = (2, 4, 0)
    _required_framework_version = (2, 0, 0)
    deleted = "(deleted)"
    smear = "<potentially smeared>"

    framework.require_interface_version(*_required_framework_version)

    @classmethod
    def _get_path_file(cls, task, filp) -> str:
        """Returns the file pathname relative to the task's root directory.

        Args:
            task (task_struct): A reference task
            filp (file *): A pointer to an open file

        Returns:
            str: File pathname relative to the task's root directory.
        """
        rdentry = task.fs.get_root_dentry()
        rmnt = task.fs.get_root_mnt()
        vfsmnt = filp.get_vfsmnt()
        dentry = filp.get_dentry()

        return cls.do_get_path(rdentry, rmnt, dentry, vfsmnt)

    @classmethod
    def get_path_mnt(cls, task, mnt) -> str:
        """Returns the mount point pathname relative to the task's root directory.

        Args:
            task (task_struct): A reference task
            mnt (vfsmount or mount): A mounted filesystem or a mount point.
                - kernels < 3.3 type is 'vfsmount'
                - kernels >= 3.3 type is 'mount'

        Returns:
            str: Pathname of the mount point relative to the task's root directory.
        """
        rdentry = task.fs.get_root_dentry()
        rmnt = task.fs.get_root_mnt()

        vfsmnt = mnt.get_vfsmnt_current()
        dentry = mnt.get_dentry_current()

        return cls.do_get_path(rdentry, rmnt, dentry, vfsmnt)

    @classmethod
    def do_get_path(cls, rdentry, rmnt, dentry, vfsmnt) -> Union[None, str]:
        """Returns a pathname of the mount point or file
        It mimics the Linux kernel prepend_path function.

        Args:
            rdentry (dentry *): A pointer to the root dentry
            rmnt (vfsmount *): A pointer to the root vfsmount
            dentry (dentry *): A pointer to the dentry
            vfsmnt (vfsmount/vfsmount *): A vfsmount object (kernels >= 3.3) or a
                vfsmount pointer (kernels < 3.3)

        Returns:
            str: Pathname of the mount point or file
        """

        if not (rdentry and rdentry.is_readable() and rmnt and rmnt.is_readable()):
            return ""

        if isinstance(vfsmnt, objects.Pointer) and not (
            vfsmnt and vfsmnt.is_readable()
        ):
            # vfsmnt can be the vfsmount object itself (>=3.3) or a vfsmount * (<3.3)
            return ""

        inode = dentry.d_inode
        path_reversed = []
        smeared = False
        while (
            dentry
            and dentry.is_readable()
            and (dentry != rdentry or not vfsmnt.is_equal(rmnt))
        ):
            if dentry == vfsmnt.get_mnt_root() or dentry.is_root():
                # Escaped?
                if dentry != vfsmnt.get_mnt_root():
                    break

                # Global root?
                if not vfsmnt.has_parent():
                    break

                dentry = vfsmnt.get_dentry_parent()
                vfsmnt = vfsmnt.get_vfsmnt_parent()

                continue

            parent = dentry.d_parent
            dname = dentry.d_name.name_as_str()

            # empty dentry names are most likely
            # the result of smearing
            if not dname:
                smeared = True
            path_reversed.append(dname.strip("/"))
            dentry = parent

        path = "/" + "/".join(reversed(path_reversed))
        if smeared:
            # if there is smear the missing dname will be empty. e.g. if the normal
            # path would be /foo/bar/baz, but bar is missing due to smear the results
            # returned here will show /foo//baz. Note the // for the missing dname.
            return f"{LinuxUtilities.smear} {path}"

        if inode and inode.is_readable() and inode.is_valid() and inode.i_nlink == 0:
            path = f" {path} {LinuxUtilities.deleted}"
        return path

    @classmethod
    def _get_new_sock_pipe_path(cls, context, task, filp) -> str:
        """Returns the sock pipe pathname relative to the task's root directory.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            task (task_struct): A reference task
            filp (file *): A pointer to a sock pipe open file

        Returns:
            str: Sock pipe pathname relative to the task's root directory.
        """
        # FIXME: This function must be moved to the 'dentry' object extension
        # Also, the scope of this function went beyond the sock pipe path, so we need to rename this.
        # Once https://github.com/volatilityfoundation/volatility3/pull/1263 is merged, replace the
        # dentry inode getters

        if not (filp and filp.is_readable()):
            return f"<invalid file pointer> {filp:x}"

        dentry = filp.get_dentry()
        if not (dentry and dentry.is_readable()):
            return f"<invalid dentry pointer> {dentry:x}"

        kernel_module = cls.get_module_from_volobj_type(context, dentry)

        sym_addr = dentry.d_op.d_dname
        if not (sym_addr and sym_addr.is_readable()):
            return f"<invalid d_dname pointer> {sym_addr:x}"

        symbs = list(kernel_module.get_symbols_by_absolute_location(sym_addr))

        inode = dentry.d_inode
        if not (inode and inode.is_readable() and inode.is_valid()):
            return f"<invalid dentry inode> {inode:x}"

        if len(symbs) == 1:
            sym = symbs[0].split(constants.BANG)[1]

            if sym == "sockfs_dname":
                pre_name = "socket"

            elif sym == "anon_inodefs_dname":
                pre_name = "anon_inode"

            elif sym == "pipefs_dname":
                pre_name = "pipe"

            elif sym == "simple_dname":
                name = dentry.d_name.name
                if name:
                    pre_name = name.dereference().cast(
                        "string", max_length=255, errors="replace"
                    )
                    return "/" + pre_name + " (deleted)"
                else:
                    pre_name = ""

            elif sym == "ns_dname":
                # From Kernels 3.19

                # In Kernels >= 6.9, see Linux kernel commit 1fa08aece42512be072351f482096d5796edf7ca
                # ns_common->stashed change from 'atomic64_t' to 'dentry*'
                try:
                    ns_common_type = kernel_module.get_type("ns_common")
                    stashed_template = ns_common_type.child_template("stashed")
                    stashed_type_full_name = stashed_template.vol.type_name
                    stashed_type_name = stashed_type_full_name.split(constants.BANG)[1]
                    if stashed_type_name == "atomic64_t":
                        # 3.19 <= Kernels < 6.9
                        fsdata_ptr = dentry.d_fsdata
                        if not (fsdata_ptr and fsdata_ptr.is_readable()):
                            raise IndexError

                        ns_ops = fsdata_ptr.dereference().cast("proc_ns_operations")
                    else:
                        # Kernels >= 6.9
                        private_ptr = inode.i_private
                        if not (private_ptr and private_ptr.is_readable()):
                            raise IndexError

                        ns_common = private_ptr.dereference().cast("ns_common")
                        ns_ops = ns_common.ops

                    pre_name = utility.pointer_to_string(ns_ops.name, 255)
                except (exceptions.SymbolError, IndexError):
                    pre_name = "<unsupported ns_dname implementation>"
            else:
                pre_name = f"<unsupported d_op symbol> {sym}"
        else:
            pre_name = f"<unknown d_dname pointer> {sym_addr:x}"

        return f"{pre_name}:[{inode.i_ino:d}]"

    @classmethod
    def path_for_file(cls, context, task, filp, files_only=False) -> str:
        """Returns a file (or sock pipe) pathname relative to the task's root directory.

        A 'file' structure doesn't have enough information to properly restore its
        full path we need the root mount information from task_struct to determine this

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            task (task_struct): A reference task
            filp (file *): A pointer to an open file

        Returns:
            str: A file (or sock pipe) pathname relative to the task's root directory.
        """

        # Memory smear protection: Check that both the file and dentry pointers are valid.
        try:
            dentry = filp.get_dentry()
            dentry.is_root()
        except exceptions.InvalidAddressException:
            return ""

        if dentry == 0:
            return ""

        dname_is_valid = False

        # TODO COMPARE THIS IN LSOF OUTPUT TO VOL2
        try:
            if (
                dentry.d_op
                and dentry.d_op.has_member("d_dname")
                and dentry.d_op.d_dname
            ):
                dname_is_valid = True

        except exceptions.InvalidAddressException:
            dname_is_valid = False

        if dname_is_valid and not files_only:
            ret = LinuxUtilities._get_new_sock_pipe_path(context, task, filp)
        else:
            ret = LinuxUtilities._get_path_file(task, filp)

        return ret

    @classmethod
    def files_descriptors_for_process(
        cls,
        context: interfaces.context.ContextInterface,
        symbol_table: str,
        task: interfaces.objects.ObjectInterface,
        files_only: bool = False,
    ):
        try:
            files = task.files
            fd_table = files.get_fds()
            if fd_table == 0:
                return None

            max_fds = files.get_max_fds()
        except exceptions.InvalidAddressException:
            return None

        # corruption check
        if max_fds > 500000:
            return None

        file_type = symbol_table + constants.BANG + "file"

        fds = objects.utility.array_of_pointers(
            fd_table, count=max_fds, subtype=file_type, context=context
        )

        for fd_num, filp in enumerate(fds):
            if filp and filp.is_readable():
                full_path = LinuxUtilities.path_for_file(
                    context, task, filp, files_only
                )

                yield fd_num, filp, full_path

    @classmethod
    @deprecation.method_being_removed(
        removal_date="2025-09-25",
        message="Callers to this method should adapt `linux_utilities_modules.Modules.run_module_scanners`",
    )
    def mask_mods_list(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        mods: Iterator[interfaces.objects.ObjectInterface],
    ) -> List[Tuple[str, int, int]]:
        """
        DEPRECATED: use "volatility3.framework.symbols.linux.utilities.modules.Modules.mask_mods_list" instead.

        A helper function to mask the starting and end address of kernel modules
        """
        return linux_utilities_modules.Modules.mask_mods_list(context, layer_name, mods)

    @classmethod
    @deprecation.method_being_removed(
        removal_date="2025-09-25",
        message="Callers to this method should adapt `linux_utilities_modules.Modules.run_module_scanners`",
    )
    def generate_kernel_handler_info(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        mods_list: Iterator[interfaces.objects.ObjectInterface],
    ) -> List[Tuple[str, int, int]]:
        """
        This method is being deprecated. Use `linux_utilities_modules.Modules.run_module_scanners` to map kernel pointers to modules")

        A helper function that gets the beginning and end address of the kernel module
        """

        kernel = context.modules[kernel_module_name]

        mask = context.layers[kernel.layer_name].address_mask

        start_addr = kernel.object_from_symbol("_text")
        start_addr = start_addr.vol.offset & mask

        end_addr = kernel.object_from_symbol("_etext")
        end_addr = end_addr.vol.offset & mask

        return [
            (constants.linux.KERNEL_NAME, start_addr, end_addr)
        ] + linux_utilities_modules.Modules.mask_mods_list(
            context, kernel.layer_name, mods_list
        )

    @classmethod
    @deprecation.deprecated_method(
        replacement=linux_utilities_modules.Modules.lookup_module_address,
        removal_date="2025-09-25",
        replacement_version=(2, 0, 0),
    )
    def lookup_module_address(
        cls,
        kernel_module: interfaces.context.ModuleInterface,
        handlers: List[Tuple[str, int, int]],
        target_address: int,
    ) -> Tuple[str, str]:
        """
        DEPRECATED: use "volatility3.framework.symbols.linux.utilities.modules.Modules.lookup_module_address" instead.

        Searches between the start and end address of the kernel module using target_address.
        Returns the module and symbol name of the address provided.
        """
        return linux_utilities_modules.Modules.lookup_module_address(
            kernel_module.context, kernel_module.name, handlers, target_address
        )

    @classmethod
    def walk_internal_list(
        cls,
        vmlinux: interfaces.context.ModuleInterface,
        struct_name: str,
        list_member: str,
        list_start: interfaces.objects.ObjectInterface,
        max_count: int = 4096,
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """
        An API that provides generic, smear-resistant enumeration of embedded lists

        Args:
            vmlinux:
            struct_name: name of the structure of the list elements
            list_member: name of the list_member holding the internal list
            list_start: Starting (head) member of the list
            max_count: Optional maximum amount of list elements that will be yielded

        Returns:
            Instances of `struct_name`
        """

        count = 0
        seen = set()

        while list_start:
            if list_start.vol.offset in seen:
                vollog.debug(
                    "walk_internal_list: Repeat entry found. Stopping enumeration"
                )
                break
            seen.add(list_start.vol.offset)

            if not (list_start and list_start.is_readable()):
                break

            list_struct = vmlinux.object(
                object_type=struct_name, offset=list_start.vol.offset, absolute=True
            )

            yield list_struct

            list_start = getattr(list_struct, list_member)

            if count == max_count:
                vollog.debug(
                    f"walk_internal_list: Breaking list enumeration at maximum allowed count of {count}"
                )
                break

            count += 1

    @classmethod
    def container_of(
        cls,
        addr: int,
        type_name: str,
        member_name: str,
        vmlinux: interfaces.context.ModuleInterface,
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Cast a member of a structure out to the containing structure.
        It mimics the Linux kernel macro container_of() see include/linux.kernel.h

        Args:
            addr: The pointer to the member.
            type_name: The type of the container struct this is embedded in.
            member_name: The name of the member within the struct.
            vmlinux: The kernel symbols object

        Returns:
            The constructed object or None
        """

        if not addr:
            return None

        type_dec = vmlinux.get_type(type_name)
        member_offset = type_dec.relative_child_offset(member_name)
        container_addr = addr - member_offset
        layer = vmlinux.context.layers[vmlinux.layer_name]
        if not layer.is_valid(container_addr):
            return None

        return vmlinux.object(
            object_type=type_name, offset=container_addr, absolute=True
        )

    @classmethod
    def get_module_from_volobj_type(
        cls,
        context: interfaces.context.ContextInterface,
        volobj: interfaces.objects.ObjectInterface,
    ) -> interfaces.context.ModuleInterface:
        """Get the vmlinux from a vol obj

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            volobj (vol object): A vol object

        Raises:
            ValueError: If it cannot obtain any module from the symbol table

        Returns:
            A kernel object (vmlinux)
        """
        symbol_table = volobj.get_symbol_table_name()
        module_names = context.modules.get_modules_by_symbol_tables(symbol_table)
        module_names = list(module_names)

        if not module_names:
            raise ValueError(f"No module using the symbol table '{symbol_table}'")

        kernel_module_name = module_names[0]
        kernel = context.modules[kernel_module_name]

        return kernel

    @classmethod
    def convert_fourcc_code(cls, code: int) -> str:
        """Convert a fourcc integer back to its fourcc string representation.

        Args:
            code: the numerical representation of the fourcc

        Returns:
            The fourcc code string.
        """

        code_bytes_length = (code.bit_length() + 7) // 8
        return "".join(
            [chr((code >> (i * 8)) & 0xFF) for i in range(code_bytes_length)]
        )


class IDStorage(ABC):
    """Abstraction to support both XArray and RadixTree"""

    # Dynamic values, these will be initialized later
    CHUNK_SHIFT = None
    CHUNK_SIZE = None
    CHUNK_MASK = None

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ):
        self.vmlinux = context.modules[kernel_module_name]
        self.vmlinux_layer = self.vmlinux.context.layers[self.vmlinux.layer_name]

        self.pointer_size = self.vmlinux.get_type("pointer").size
        # Dynamically work out the (XA_CHUNK|RADIX_TREE_MAP)_SHIFT values based on
        # the node.slots[] array size
        node_type = self.vmlinux.get_type(self.node_type_name)
        slots_array_size = node_type.child_template("slots").count

        # Calculate the LSB index - 1
        self.CHUNK_SHIFT = slots_array_size.bit_length() - 1
        self.CHUNK_SIZE = 1 << self.CHUNK_SHIFT
        self.CHUNK_MASK = self.CHUNK_SIZE - 1

    @classmethod
    def choose_id_storage(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> "IDStorage":
        """Returns the appropriate ID storage data structure instance for the current kernel implementation.
        This is used by the IDR and the PageCache to choose between the XArray and RadixTree.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the kernel module on which to operate

        Returns:
            The appropriate ID storage instance for the current kernel
        """
        vmlinux = context.modules[kernel_module_name]
        address_space_type = vmlinux.get_type("address_space")
        address_space_has_i_pages = address_space_type.has_member("i_pages")
        i_pages_type_name = (
            address_space_type.child_template("i_pages").vol.type_name
            if address_space_has_i_pages
            else ""
        )
        i_pages_is_xarray = i_pages_type_name.endswith(constants.BANG + "xarray")
        i_pages_is_radix_tree_root = i_pages_type_name.endswith(
            constants.BANG + "radix_tree_root"
        ) and vmlinux.get_type("radix_tree_root").has_member("xa_head")

        if i_pages_is_xarray or i_pages_is_radix_tree_root:
            return XArray(context, kernel_module_name)
        else:
            return RadixTree(context, kernel_module_name)

    @property
    @abstractmethod
    def node_type_name(self) -> str:
        """Returns the Tree implementation node type name

        Returns:
            A string with the node type name
        """
        raise NotImplementedError()

    @property
    def tag_internal_value(self) -> int:
        """Returns the internal node flag for the tree"""
        raise NotImplementedError()

    @abstractmethod
    def node_is_internal(self, nodep) -> bool:
        """Checks if the node is internal"""
        raise NotImplementedError

    @abstractmethod
    def is_node_tagged(self, nodep) -> bool:
        """Checks if the node pointer is tagged"""
        raise NotImplementedError

    @abstractmethod
    def untag_node(self, nodep) -> int:
        """Untags a node pointer"""
        raise NotImplementedError

    @abstractmethod
    def get_tree_height(self, treep) -> int:
        """Returns the tree height"""
        raise NotImplementedError

    @abstractmethod
    def get_node_height(self, nodep) -> int:
        """Returns the node height"""
        raise NotImplementedError

    @abstractmethod
    def get_head_node(self, tree) -> Optional[int]:
        """Returns a pointer to the tree's head"""
        raise NotImplementedError

    @abstractmethod
    def is_valid_node(self, nodep) -> bool:
        """Validates a node pointer"""
        raise NotImplementedError

    def nodep_to_node(self, nodep) -> interfaces.objects.ObjectInterface:
        """Instantiates a tree node from its pointer

        Args:
            nodep: Pointer to the XArray/RadixTree node

        Returns:
            A XArray/RadixTree node instance
        """
        node = self.vmlinux.object(self.node_type_name, offset=nodep, absolute=True)
        return node

    def _slot_to_nodep(self, slot) -> int:
        if self.node_is_internal(slot):
            nodep = slot & ~self.tag_internal_value
        else:
            nodep = slot

        return nodep

    def _iter_node(self, nodep, height) -> Iterator[int]:
        node = self.nodep_to_node(nodep)
        node_slots = node.slots
        for off in range(self.CHUNK_SIZE):
            try:
                slot = node_slots[off]
            except exceptions.InvalidAddressException:
                continue

            if slot == 0:
                continue

            nodep = self._slot_to_nodep(slot)

            if height == 1:
                if self.is_valid_node(nodep):
                    yield nodep
            else:
                yield from self._iter_node(nodep, height - 1)

    def get_entries(self, root: interfaces.objects.ObjectInterface) -> Iterator[int]:
        """Walks the tree data structure

        Args:
            root: The tree root object

        Yields:
            A tree node pointer
        """
        height = self.get_tree_height(root.vol.offset)

        nodep = self.get_head_node(root)
        if not (nodep and nodep.is_readable()):
            return

        # Keep the internal flag before untagging it
        is_internal = self.node_is_internal(nodep)
        if self.is_node_tagged(nodep):
            nodep = self.untag_node(nodep)

        if is_internal:
            height = self.get_node_height(nodep)

        if height == 0:
            if self.is_valid_node(nodep):
                yield nodep
        else:
            yield from self._iter_node(nodep, height)


class XArray(IDStorage):
    XARRAY_TAG_MASK = 3
    XARRAY_TAG_INTERNAL = 2

    def get_tree_height(self, treep) -> int:
        return 0

    @property
    def node_type_name(self) -> str:
        return "xa_node"

    @property
    def tag_internal_value(self) -> int:
        return self.XARRAY_TAG_INTERNAL

    def get_node_height(self, nodep) -> int:
        node = self.nodep_to_node(nodep)
        return (node.shift // self.CHUNK_SHIFT) + 1

    def get_head_node(self, tree) -> Optional[int]:
        try:
            return tree.xa_head
        except exceptions.InvalidAddressException:
            return None

    def node_is_internal(self, nodep) -> bool:
        return (nodep & self.XARRAY_TAG_MASK) == self.XARRAY_TAG_INTERNAL

    def is_node_tagged(self, nodep) -> bool:
        return (nodep & self.XARRAY_TAG_MASK) != 0

    def untag_node(self, nodep) -> int:
        return nodep & (~self.XARRAY_TAG_MASK)

    def is_valid_node(self, nodep) -> bool:
        # It should have the tag mask clear
        return not self.is_node_tagged(nodep)


class RadixTree(IDStorage):
    RADIX_TREE_INTERNAL_NODE = 1
    RADIX_TREE_EXCEPTIONAL_ENTRY = 2
    RADIX_TREE_ENTRY_MASK = 3
    RADIX_TREE_MAP_SHIFT = 6  # CONFIG_BASE_FULL

    # Dynamic values. These will be initialized later
    RADIX_TREE_INDEX_BITS = None
    RADIX_TREE_MAX_PATH = None
    RADIX_TREE_HEIGHT_SHIFT = None
    RADIX_TREE_HEIGHT_MASK = None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        char_bits = 8
        self.RADIX_TREE_INDEX_BITS = char_bits * self.pointer_size
        self.RADIX_TREE_MAX_PATH = int(
            math.ceil(self.RADIX_TREE_INDEX_BITS / float(self.CHUNK_SHIFT))
        )
        self.RADIX_TREE_HEIGHT_SHIFT = self.RADIX_TREE_MAX_PATH + 1
        self.RADIX_TREE_HEIGHT_MASK = (1 << self.RADIX_TREE_HEIGHT_SHIFT) - 1

        if not self.vmlinux.has_type("radix_tree_root"):
            # In kernels 4.20, RADIX_TREE_INTERNAL_NODE flag took RADIX_TREE_EXCEPTIONAL_ENTRY's
            # value. RADIX_TREE_EXCEPTIONAL_ENTRY was removed but that's managed in is_valid_node()
            # Note that the Radix Tree is still in use for IDR, even after kernels 4.20 when XArray
            # mostly replace it
            self.RADIX_TREE_INTERNAL_NODE = 2

    @property
    def node_type_name(self) -> str:
        return "radix_tree_node"

    @property
    def tag_internal_value(self) -> int:
        return self.RADIX_TREE_INTERNAL_NODE

    def get_tree_height(self, treep) -> int:
        with contextlib.suppress(exceptions.SymbolError):
            if self.vmlinux.get_type("radix_tree_root").has_member("height"):
                # kernels < 4.7 d0891265bbc988dc91ed8580b38eb3dac128581b
                radix_tree_root = self.vmlinux.object(
                    "radix_tree_root", offset=treep, absolute=True
                )
                return radix_tree_root.height

        # kernels >= 4.7
        return 0

    @functools.cached_property
    def _max_height_array(self):
        if self.vmlinux.has_symbol("height_to_maxindex"):
            # 2.6.24 26fb1589cb0aaec3a0b4418c54f30c1a2b1781f6 <= Kernels < 4.7 d0891265bbc988dc91ed8580b38eb3dac128581b
            return self.vmlinux.object_from_symbol("height_to_maxindex")
        elif self.vmlinux.has_symbol("height_to_maxnodes"):
            # 4.8 c78c66d1ddfdbd2353f3fcfeba0268524537b096 <= kernels < 4.20 8cf2f98411e3a0865026a1061af637161b16d32b
            return self.vmlinux.object_from_symbol("height_to_maxnodes")

        return None

    def _radix_tree_maxindex(self, node, height) -> int:
        """Return the maximum key which can be store into a radix tree with this height."""

        if self._max_height_array:
            # 2.6.24 <= kernels <= 4.20 See _max_height_array()
            return self._max_height_array[height]
        else:
            # Kernels >= 4.20
            return (self.CHUNK_SIZE << node.shift) - 1

    def get_node_height(self, nodep) -> int:
        node = self.nodep_to_node(nodep)
        if hasattr(node, "shift"):
            # 4.7 <= Kernels < 4.20
            height = (node.shift // self.CHUNK_SHIFT) + 1
        elif hasattr(node, "path"):
            # 3.15 <= Kernels < 4.7
            height = node.path & self.RADIX_TREE_HEIGHT_MASK
        elif hasattr(node, "height"):
            # Kernels < 3.15
            height = node.height
        else:
            raise exceptions.VolatilityException("Cannot find radix-tree node height")

        if self._max_height_array and not (0 <= height < self._max_height_array.count):
            error_msg = f"Radix Tree node {node.vol.offset:#x} height {height} exceeds max height of {self._max_height_array.count}"
            vollog.error(error_msg)
            raise exceptions.LinuxPageCacheException(error_msg)

        return height

    def get_head_node(self, tree) -> Optional[int]:
        try:
            return tree.rnode
        except exceptions.InvalidAddressException:
            return None

    def node_is_internal(self, nodep) -> bool:
        return (nodep & self.RADIX_TREE_INTERNAL_NODE) != 0

    def is_node_tagged(self, nodep) -> bool:
        return self.node_is_internal(nodep)

    def untag_node(self, nodep) -> int:
        return nodep & (~self.RADIX_TREE_ENTRY_MASK)

    def _is_exceptional_node(self, nodep) -> bool:
        # In kernels 4.20, exceptional nodes were removed and internal entries took their bitmask
        return (
            self.vmlinux.has_type("radix_tree_root")
            and (nodep & self.RADIX_TREE_ENTRY_MASK)
            == self.RADIX_TREE_EXCEPTIONAL_ENTRY
        )

    def is_valid_node(self, nodep) -> bool:
        return not self._is_exceptional_node(nodep)


class PageCache:
    """Linux Page Cache abstraction"""

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        page_cache: interfaces.objects.ObjectInterface,
    ):
        """
        Args:
            context: interfaces.context.ContextInterface,
            kernel_module_name: The name of the kernel module on which to operate
            page_cache: Page cache address space
        """
        self.vmlinux = context.modules[kernel_module_name]

        self._page_cache = page_cache
        self._idstorage = IDStorage.choose_id_storage(context, kernel_module_name)

    def get_cached_pages(self) -> Iterator[interfaces.objects.ObjectInterface]:
        """Returns all page cache contents

        Yields:
            Page objects
        """
        layer = self.vmlinux.context.layers[self.vmlinux.layer_name]
        for page_addr in self._idstorage.get_entries(self._page_cache.i_pages):
            if not layer.is_valid(page_addr):
                error_msg = f"Invalid cached page address at {page_addr:#x}, aborting"
                vollog.error(error_msg)
                raise exceptions.LinuxPageCacheException(error_msg)

            page = self.vmlinux.object("page", offset=page_addr, absolute=True)
            if not page.is_valid():
                error_msg = f"Invalid cached page at {page_addr:#x}, aborting"
                vollog.error(error_msg)
                raise exceptions.LinuxPageCacheException(error_msg)

            yield page


class VMCoreInfo(interfaces.configuration.VersionableInterface):
    _required_framework_version = (2, 11, 0)

    _version = (1, 0, 0)

    @classmethod
    def _vmcoreinfo_data_to_dict(
        cls,
        vmcoreinfo_data,
    ) -> Optional[Dict[str, str]]:
        """Converts the input VMCoreInfo data buffer into a dictionary"""

        # Ensure the whole buffer is printable
        printable_bytes_set = set(string.printable.encode())
        if not all(byte in printable_bytes_set for byte in vmcoreinfo_data):
            # Abort, we are in the wrong place
            return None

        vmcoreinfo_dict = dict()
        for line in vmcoreinfo_data.decode().splitlines():
            if not line:
                break

            key, value = line.split("=", 1)
            vmcoreinfo_dict[key] = cls._parse_value(key, value)

        return vmcoreinfo_dict

    @classmethod
    def _parse_value(cls, key, value):
        if key.startswith("SYMBOL(") or key == "KERNELOFFSET":
            return int(value, 16)
        elif key.startswith(("NUMBER(", "LENGTH(", "SIZE(", "OFFSET(")):
            return int(value, 0)
        elif key == "PAGESIZE":
            return int(value, 0)

        # Default, as string
        return value

    @classmethod
    def search_vmcoreinfo_elf_note(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Iterator[Tuple[int, Dict[str, str]]]:
        """Enumerates each VMCoreInfo ELF note table found in memory along with its offset.

        This approach is independent of any external ISF symbol or type, requiring only the
        Elf64_Note found in 'elf.json', which is already included in the framework.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            layer_name: The layer within the context in which the module exists
            progress_callback: A function that takes a percentage (and an optional description) that will be called periodically

        Yields:
            Tuples with the VMCoreInfo ELF note offset and the VMCoreInfo table parsed in a dictionary.
        """

        elf_table_name = intermed.IntermediateSymbolTable.create(
            context, "elf_symbol_table", "linux", "elf"
        )
        module = context.module(elf_table_name, layer_name, 0)
        layer = context.layers[layer_name]

        # Both Elf32_Note and Elf64_Note are of the same size
        elf_note_size = context.symbol_space[elf_table_name].get_type("Elf64_Note").size

        for vmcoreinfo_offset in layer.scan(
            scanner=scanners.BytesScanner(linux_constants.VMCOREINFO_MAGIC_ALIGNED),
            context=context,
            progress_callback=progress_callback,
        ):
            # vmcoreinfo_note kernels >= 2.6.24 fd59d231f81cb02870b9cf15f456a897f3669b4e
            vmcoreinfo_elf_note_offset = vmcoreinfo_offset - elf_note_size

            # Elf32_Note and Elf64_Note are identical, so either can be used interchangeably here
            elf_note = module.object(
                object_type="Elf64_Note",
                offset=vmcoreinfo_elf_note_offset,
                absolute=True,
            )

            # Ensure that we are within an ELF note
            if (
                elf_note.n_namesz != len(linux_constants.VMCOREINFO_MAGIC)
                or elf_note.n_type != 0
                or elf_note.n_descsz == 0
            ):
                continue

            vmcoreinfo_data_offset = vmcoreinfo_offset + len(
                linux_constants.VMCOREINFO_MAGIC_ALIGNED
            )

            # Also, confirm this with the first tag, which has consistently been OSRELEASE
            vmcoreinfo_data = layer.read(vmcoreinfo_data_offset, elf_note.n_descsz)
            if not vmcoreinfo_data.startswith(linux_constants.OSRELEASE_TAG):
                continue

            table = cls._vmcoreinfo_data_to_dict(vmcoreinfo_data)
            if not table:
                # Wrong VMCoreInfo note offset, keep trying
                continue

            # A valid VMCoreInfo ELF note exists at 'vmcoreinfo_elf_note_offset'
            yield vmcoreinfo_elf_note_offset, table



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0

import json
import logging
import os
from typing import List

from volatility3.framework import renderers, interfaces, objects, exceptions, constants
from volatility3.framework.configuration import requirements
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)


class Privs(interfaces.plugins.PluginInterface):
    """Lists process token privileges"""

    _version = (1, 2, 0)
    _required_framework_version = (2, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Find the sids json path (or raise error if its not in the plugin directory).
        for plugin_dir in constants.PLUGINS_PATH:
            sids_json_file_name = os.path.join(
                plugin_dir, os.path.join("windows", "sids_and_privileges.json")
            )
            if os.path.exists(sids_json_file_name):
                break
        else:
            vollog.log(
                constants.LOGLEVEL_VVV,
                "sids_and_privileges.json file is missing plugin error",
            )
            raise RuntimeError(
                "The sids_and_privileges.json file missed from you plugin directory"
            )

        # Get service sids dictionary (we need only the service sids).
        with open(sids_json_file_name) as file_handle:
            temp_json = json.load(file_handle)["privileges"]
            self.privilege_info = {
                int(priv_num): temp_json[priv_num] for priv_num in temp_json
            }

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
        ]

    def _generator(self, procs):
        for task in procs:
            try:
                process_token = task.Token.dereference().cast("_TOKEN")
            except exceptions.InvalidAddressException:
                vollog.log(constants.LOGLEVEL_VVV, "Skip invalid token.")
                continue

            for value, present, enabled, default in process_token.privileges():
                # Skip privileges whose bit positions cannot be
                # translated to a privilege name
                if not self.privilege_info.get(int(value)):
                    vollog.log(
                        constants.LOGLEVEL_VVV, f"Skip invalid privilege ({value})."
                    )
                    continue

                name, desc = self.privilege_info.get(int(value))

                # Set the attributes
                attributes = []
                if present:
                    attributes.append("Present")
                if enabled:
                    attributes.append("Enabled")
                if default:
                    attributes.append("Default")

                yield (
                    0,
                    [
                        int(task.UniqueProcessId),
                        objects.utility.array_to_string(task.ImageFileName),
                        int(value),
                        str(name),
                        ",".join(attributes),
                        str(desc),
                    ],
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Value", int),
                ("Privilege", str),
                ("Attributes", str),
                ("Description", str),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=filter_func,
                )
            ),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import dataclasses
import datetime
import enum
import itertools
import logging
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, Union

from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import registry
from volatility3.framework.renderers import conversion
from volatility3.framework.symbols.windows.extensions import registry as reg_extensions
from volatility3.plugins import timeliner
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)

#######################################################################
# More information about the following enums can be found in the report
# 'Analysis of the AmCache` by Blanche Lagny, 2019
#######################################################################


class Win8FileValName(enum.Enum):
    """
    An enumeration that creates a helpful mapping of opaque Windows 8 Amcache
    'File' subkey value names to their human-readable equivalent.
    """

    ProgramID = "100"
    SHA1Hash = "101"
    Product = "0"
    Company = "1"
    Size = "6"
    SizeOfImage = "7"
    PEHeaderChecksum = "9"
    LastModTime = "11"  # REG_QWORD FILETIME
    CreateTime = "12"  # REG_QWORD FILETIME
    Path = "15"
    LastModTime2 = "17"  # REG_QWORD FILETIME
    Version = "d"
    CompileTime = "f"  # REG_QWORD UNIX EPOCH


class Win8ProgramValName(enum.Enum):
    """
    An enumeration that creates a helpful mapping of opaque Windows 8 Amcache
    'Program' subkey value names to their human-readable equivalent.
    """

    Product = "0"
    Version = "1"
    Publisher = "2"
    InstallTime = "a"
    MSIProductCode = "11"
    MSIPackageCode = "12"
    ProductCode = "f"
    PackageCode = "10"


class Win10InvAppFileValName(enum.Enum):
    """
    An enumeration containing the most useful Windows 10 Amcache
    'InventoryApplicationFile' subkey value names.
    """

    FileId = "FileId"
    LinkDate = "LinkDate"
    LowerCaseLongPath = "LowerCaseLongPath"
    ProductName = "ProductName"
    ProductVersion = "ProductVersion"
    ProgramID = "ProgramId"
    Publisher = "Publisher"


class Win10InvAppValName(enum.Enum):
    """
    An enumeration containing the most useful Windows 10 Amcache
    'InventoryApplication' subkey value names.
    """

    InstallDate = "InstallDate"
    Name = "Name"
    Publisher = "Publisher"
    RootDirPath = "RootDirPath"
    Version = "Version"


class Win10DriverBinaryValName(enum.Enum):
    """
    An enumeration containing the most useful Windows 10 Amcache
    'InventoryDriverBinary' subkey value names.
    """

    DriverId = "DriverId"
    DriverName = "DriverName"
    DriverCompany = "DriverCompany"
    Product = "Product"
    Service = "Service"
    DriverTimeStamp = "DriverTimeStamp"


class AmcacheEntryType(enum.IntEnum):
    Driver = 1
    Program = 2
    File = 3


NullableString = Union[str, None, interfaces.renderers.BaseAbsentValue]
NullableDatetime = Union[datetime.datetime, None, interfaces.renderers.BaseAbsentValue]


@dataclasses.dataclass
class _AmcacheEntry:
    """
    A class containing all information about an entry from the Amcache registry hive.
    Because all values could potentially be paged out of memory or malformed, they are all
    a union between their expected value and `interfaces.renderers.BaseAbsentValue`.
    """

    entry_type: str
    path: NullableString = renderers.NotApplicableValue()
    company: NullableString = renderers.NotApplicableValue()
    last_modify_time: NullableDatetime = renderers.NotApplicableValue()
    last_modify_time_2: NullableDatetime = renderers.NotApplicableValue()
    install_time: NullableDatetime = renderers.NotApplicableValue()
    compile_time: NullableDatetime = renderers.NotApplicableValue()
    sha1_hash: NullableString = renderers.NotApplicableValue()
    service: NullableString = renderers.NotApplicableValue()
    product_name: NullableString = renderers.NotApplicableValue()
    product_version: NullableString = renderers.NotApplicableValue()


def _entry_sort_key(entry_tuple: Tuple[NullableString, _AmcacheEntry]) -> str:
    """Sorts entries by program ID. This is broken out as a function here
    to ensure consistency in sorting between the `group_by` and `sorted` function
    invocations.
    """
    program_id, _ = entry_tuple
    key = program_id if isinstance(program_id, str) else ""
    return key


def _get_string_value(
    values: Dict[str, reg_extensions.CM_KEY_VALUE], name: str
) -> NullableString:
    try:
        value = values[name]
    except KeyError:
        return renderers.NotAvailableValue()

    data = value.decode_data()
    if not isinstance(data, bytes):
        return renderers.UnparsableValue()

    return data.decode("utf-16le", errors="replace").rstrip("\u0000")


def _get_datetime_filetime_value(
    values: Dict[str, reg_extensions.CM_KEY_VALUE], name: str
) -> NullableDatetime:
    try:
        value = values[name]
    except KeyError:
        return renderers.NotAvailableValue()

    data = value.decode_data()
    if not isinstance(data, int):
        return renderers.UnparsableValue()

    return conversion.wintime_to_datetime(data)


def _get_datetime_utc_epoch_value(
    values: Dict[str, reg_extensions.CM_KEY_VALUE], name: str
) -> NullableDatetime:
    try:
        value = values[name]
    except KeyError:
        return renderers.NotAvailableValue()

    data = value.decode_data()
    if not isinstance(data, (int, float)):
        return renderers.UnparsableValue()

    try:
        return datetime.datetime.fromtimestamp(float(data), datetime.timezone.utc)
    except (ValueError, OverflowError, OSError):
        return renderers.UnparsableValue()


def _get_datetime_str_value(
    values: Dict[str, reg_extensions.CM_KEY_VALUE], name: str
) -> NullableDatetime:
    try:
        value = values[name]
    except KeyError:
        return renderers.NotAvailableValue()

    data = value.decode_data()
    if not isinstance(data, int):
        return renderers.UnparsableValue()

    if isinstance(data, str):
        try:
            return datetime.datetime.strptime(data, "%m/%d/%Y %H:%M:%S")
        except ValueError:
            return renderers.UnparsableValue()
    else:
        return renderers.UnparsableValue()


class Amcache(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Extract information on executed applications from the AmCache."""

    _required_framework_version = (2, 0, 0)

    # 2.0.0 - changed the signature of get_amcache_hive
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
        ]

    def generate_timeline(
        self,
    ) -> Iterator[Tuple[str, timeliner.TimeLinerType, datetime.datetime]]:
        for _, entry in self._generator():
            if isinstance(entry.last_modify_time, datetime.datetime):
                yield (
                    f"Amcache: {entry.entry_type} {entry.path} registry key modified",
                    timeliner.TimeLinerType.MODIFIED,
                    entry.last_modify_time,
                )
            if isinstance(entry.last_modify_time_2, datetime.datetime):
                yield (
                    f"Amcache: {entry.entry_type} {entry.path} STANDARD_INFORMATION create time",
                    timeliner.TimeLinerType.CREATED,
                    entry.last_modify_time_2,
                )
            if isinstance(entry.install_time, datetime.datetime):
                yield (
                    f"Amcache: {entry.entry_type} {entry.path} installed",
                    timeliner.TimeLinerType.CREATED,
                    entry.install_time,
                )
            if isinstance(entry.compile_time, datetime.datetime):
                yield (
                    f"Amcache: {entry.entry_type} {entry.path} compiled (PE metadata)",
                    timeliner.TimeLinerType.MODIFIED,
                    entry.compile_time,
                )

    @classmethod
    def get_amcache_hive(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
    ) -> Optional[registry.RegistryHive]:
        """Retrieves the `Amcache.hve` registry hive from the kernel module, if it can be located."""
        return next(
            hivelist.HiveList.list_hives(
                context=context,
                base_config_path=interfaces.configuration.path_join(
                    config_path, "hivelist"
                ),
                kernel_module_name=kernel_module_name,
                filter_string="amcache",
            ),
            None,
        )

    @classmethod
    def parse_file_key(
        cls, file_key: reg_extensions.CM_KEY_NODE
    ) -> Iterator[Tuple[NullableString, _AmcacheEntry]]:
        """Parses File entries from the Windows 8 `Root\\File` key.

        :param programs_key: The `Root\\File` registry key.

        :return: An iterator of tuples, where the first member is the program ID string for
        correlating `Root\\Program` entries, and the second member is the `AmcacheEntry`.
        """

        val_enum = Win8FileValName

        wanted_values = [key.value for key in val_enum]

        for file_entry_key in itertools.chain(
            *(key.get_subkeys() for key in file_key.get_subkeys())
        ):
            vollog.debug(f"Checking Win8 File key {file_entry_key.get_name()}")
            values = {
                str(value.get_name()): value
                for value in file_entry_key.get_values()
                if value.get_name() in wanted_values
            }

            program_id = _get_string_value(values, val_enum.ProgramID.value)
            path = _get_string_value(values, val_enum.Path.value)
            company = _get_string_value(values, val_enum.Company.value)
            last_mod_time = _get_datetime_filetime_value(
                values, val_enum.LastModTime.value
            )
            last_mod_time_2 = _get_datetime_filetime_value(
                values, val_enum.LastModTime2.value
            )
            install_time = _get_datetime_filetime_value(
                values, val_enum.CreateTime.value
            )
            compile_time = _get_datetime_utc_epoch_value(
                values, val_enum.CompileTime.value
            )
            sha1_hash = _get_string_value(values, val_enum.SHA1Hash.value)
            vollog.debug(f"Found sha1hash {sha1_hash}")
            product_name = _get_string_value(values, val_enum.Product.value)

            yield (
                program_id,
                _AmcacheEntry(
                    AmcacheEntryType.File.name,
                    path=path,
                    company=company,
                    last_modify_time=last_mod_time,
                    last_modify_time_2=last_mod_time_2,
                    install_time=install_time,
                    compile_time=compile_time,
                    sha1_hash=(
                        sha1_hash.lstrip("0000")
                        if isinstance(sha1_hash, str)
                        else sha1_hash
                    ),
                    product_name=product_name,
                ),
            )

    @classmethod
    def parse_programs_key(
        cls, programs_key: reg_extensions.CM_KEY_NODE
    ) -> Iterator[Tuple[str, _AmcacheEntry]]:
        """Parses Program entries from the Windows 8 `Root\\Programs` key.

        :param programs_key: The `Root\\Programs` registry key.

        :return: An iterator of tuples, where the first member is the program ID string for
        correlating `Root\\File` entries, and the second member is the `AmcacheEntry`.
        """
        val_enum = Win8ProgramValName

        wanted_values = [key.value for key in val_enum]
        for program_key in programs_key.get_subkeys():
            values = {
                str(value.get_name()): value
                for value in program_key.get_values()
                if value.get_name() in wanted_values
            }
            vollog.debug(f"Parsing Win8 Program key {program_key.get_name()}")
            program_id = program_key.get_name().strip().strip("\u0000")

            product = _get_string_value(values, val_enum.Product.value)
            company = _get_string_value(values, val_enum.Publisher.value)
            install_time = _get_datetime_utc_epoch_value(
                values, val_enum.InstallTime.value
            )
            version = _get_string_value(values, val_enum.Version.value)

            yield (
                program_id,
                _AmcacheEntry(
                    AmcacheEntryType.Program.name,
                    company=company,
                    last_modify_time=conversion.wintime_to_datetime(
                        program_key.LastWriteTime.QuadPart
                    ),
                    install_time=install_time,
                    product_name=product,
                    product_version=version,
                ),
            )

    @classmethod
    def parse_inventory_app_key(
        cls, inv_app_key: reg_extensions.CM_KEY_NODE
    ) -> Iterator[Tuple[str, _AmcacheEntry]]:
        """Parses InventoryApplication entries from the Windows 10 `Root\\InventoryApplication` key.

        :param programs_key: The `Root\\InventoryApplication` registry key.

        :return: An iterator of tuples, where the first member is the program ID string for
        correlating `Root\\InventoryApplicationFile` entries, and the second member is the `AmcacheEntry`.
        """
        val_enum = Win10InvAppValName

        wanted_values = [key.value for key in val_enum]

        for program_key in inv_app_key.get_subkeys():
            program_id = program_key.get_name()

            values = {
                str(value.get_name()): value
                for value in program_key.get_values()
                if value.get_name() in wanted_values
            }

            name = _get_string_value(values, val_enum.Name.value)
            version = _get_string_value(values, val_enum.Version.value)
            publisher = _get_string_value(values, val_enum.Publisher.value)
            path = _get_string_value(values, val_enum.RootDirPath.value)
            install_date = _get_datetime_str_value(values, val_enum.InstallDate.value)
            last_mod = conversion.wintime_to_datetime(
                program_key.LastWriteTime.QuadPart
            )

            product: str = name if isinstance(name, str) else "UNKNOWN"  # type: ignore

            yield (
                program_id.strip().strip("\u0000"),
                _AmcacheEntry(
                    AmcacheEntryType.Program.name,
                    path=path,
                    last_modify_time=last_mod,
                    install_time=install_date,
                    product_name=product,
                    company=publisher,
                    product_version=version,
                ),
            )

    @classmethod
    def parse_inventory_app_file_key(
        cls, inv_app_file_key: reg_extensions.CM_KEY_NODE
    ) -> Iterator[Tuple[NullableString, _AmcacheEntry]]:
        """Parses executable file entries from the `Root\\InventoryApplicationFile` registry key.

        :param inv_app_file_key: The `Root\\InventoryApplicationFile` registry key.
        :return: An iterator of tuples, where the first member is the program ID string for correlating
        with it's parent `InventoryApplication` program entry, and the second member is the `Amcache` entry.
        """

        val_enum = Win10InvAppFileValName

        wanted_values = [key.value for key in val_enum]

        for file_key in inv_app_file_key.get_subkeys():
            vollog.debug(
                f"Parsing Win10 InventoryApplicationFile key {file_key.get_name()}"
            )

            values = {
                str(value.get_name()): value
                for value in file_key.get_values()
                if value.get_name() in wanted_values
            }

            last_mod = conversion.wintime_to_datetime(file_key.LastWriteTime.QuadPart)
            path = _get_string_value(values, val_enum.LowerCaseLongPath.value)
            linkdate = _get_datetime_str_value(values, val_enum.LinkDate.value)
            sha1_hash = _get_string_value(values, val_enum.FileId.value)
            publisher = _get_string_value(values, val_enum.Publisher.value)
            prod_name = _get_string_value(values, val_enum.ProductName.value)
            prod_ver = _get_string_value(values, val_enum.ProductVersion.value)
            program_id = _get_string_value(values, val_enum.ProgramID.value)

            yield (
                program_id,
                _AmcacheEntry(
                    AmcacheEntryType.File.name,
                    path=path,
                    company=publisher,
                    last_modify_time=last_mod,
                    compile_time=linkdate,
                    sha1_hash=(
                        sha1_hash.lstrip("0000")
                        if isinstance(sha1_hash, str)
                        else sha1_hash
                    ),
                    product_name=prod_name,
                    product_version=prod_ver,
                ),
            )

    @classmethod
    def parse_driver_binary_key(
        cls, driver_binary_key: reg_extensions.CM_KEY_NODE
    ) -> Iterator[_AmcacheEntry]:
        """Parses information about installed drivers from the `Root\\InventoryDriverBinary` registry key.

        :param driver_binary_key: The `Root\\InventoryDriverBinary` registry key
        :return: An iterator of `AmcacheEntry`s
        """
        val_enum = Win10DriverBinaryValName

        wanted_values = [key.value for key in val_enum]

        for binary_key in driver_binary_key.get_subkeys():
            values = {
                str(value.get_name()): value
                for value in binary_key.get_values()
                if value.get_name() in wanted_values
            }

            # Depending on the Windows version, the key name will be either the name
            # of the driver, or its SHA1 hash.
            if "/" in str(binary_key.get_name()):
                driver_name = str(binary_key.get_name())
                sha1_hash = _get_string_value(values, val_enum.DriverId.name)
            else:
                sha1_hash = str(binary_key.get_name())
                driver_name = _get_string_value(values, val_enum.DriverName.name)

            if isinstance(sha1_hash, str):
                sha1_hash = sha1_hash[4:] if sha1_hash.startswith("0000") else sha1_hash

            company, product, service, last_write_time, driver_timestamp = (
                _get_string_value(values, val_enum.DriverCompany.name),
                _get_string_value(values, val_enum.Product.name),
                _get_string_value(values, val_enum.Service.name),
                conversion.wintime_to_datetime(binary_key.LastWriteTime.QuadPart),
                _get_datetime_utc_epoch_value(values, val_enum.DriverTimeStamp.name),
            )

            yield _AmcacheEntry(
                entry_type=AmcacheEntryType.Driver.name,
                path=driver_name,
                company=company,
                last_modify_time=last_write_time,
                compile_time=driver_timestamp,
                sha1_hash=(
                    sha1_hash.lstrip("0000")
                    if isinstance(sha1_hash, str)
                    else sha1_hash
                ),
                service=service,
                product_name=product,
            )

    def _generator(self) -> Iterator[Tuple[int, _AmcacheEntry]]:
        def indented(
            entry_gen: Iterable[_AmcacheEntry], indent: int = 0
        ) -> Iterator[Tuple[int, _AmcacheEntry]]:
            for item in entry_gen:
                yield indent, item

        # Building the dictionary ahead of time is much better for performance
        # vs looking up each service's DLL individually.
        amcache = self.get_amcache_hive(
            self.context, self.config_path, self.config["kernel"]
        )
        if amcache is None:
            return

        try:
            yield from indented(
                self.parse_driver_binary_key(
                    amcache.get_key("Root\\InventoryDriverBinary")  # type: ignore
                )
            )
        except (KeyError, registry.RegistryException):
            # Registry key not found
            pass

        try:
            programs: Dict[str, _AmcacheEntry] = {
                program_id: entry
                for program_id, entry in self.parse_programs_key(
                    amcache.get_key("Root\\Programs")
                )  # type: ignore
            }
        except (KeyError, registry.RegistryException):
            programs = {}

        try:
            files = sorted(
                list(
                    self.parse_file_key(amcache.get_key("Root\\File")),  # type: ignore
                ),
                key=_entry_sort_key,
            )
        except (KeyError, registry.RegistryException):
            files = []

        for program_id, file_entries in itertools.groupby(
            files,
            key=_entry_sort_key,
        ):
            files_indent = 0
            if isinstance(program_id, str):
                try:
                    program_entry = programs.pop(program_id.strip().strip("\u0000"))
                    yield (0, program_entry)

                    files_indent = 1
                except KeyError:
                    # No parent program for this file entry
                    pass
            for _, entry in file_entries:
                yield files_indent, entry

        for empty_program in programs.values():
            yield 0, empty_program

        try:
            programs: Dict[str, _AmcacheEntry] = dict(
                self.parse_inventory_app_key(
                    amcache.get_key("Root\\InventoryApplication")  # type: ignore
                )
            )
        except (KeyError, registry.RegistryException):
            programs = {}

        try:
            files = sorted(
                list(
                    self.parse_inventory_app_file_key(
                        amcache.get_key("Root\\InventoryApplicationFile")
                    ),
                    # type: ignore
                ),
                key=_entry_sort_key,
            )
        except (KeyError, registry.RegistryException):
            files = []

        for program_id, file_entries in itertools.groupby(
            files,
            key=_entry_sort_key,
        ):
            files_indent = 0

            if isinstance(program_id, str):
                try:
                    program_entry = programs.pop(program_id.strip().strip("\u0000"))
                    yield (0, program_entry)
                    files_indent = 1
                except KeyError:
                    # No parent program for this file entry
                    pass

            for _, entry in file_entries:
                yield files_indent, entry

        for empty_program in programs.values():
            yield 0, empty_program

    def run(self):
        return renderers.TreeGrid(
            [
                ("EntryType", str),
                ("Path", str),
                ("Company", str),
                ("LastModifyTime", datetime.datetime),
                ("LastModifyTime2", datetime.datetime),
                ("InstallTime", datetime.datetime),
                ("CompileTime", datetime.datetime),
                ("SHA1", str),
                ("Service", str),
                ("ProductName", str),
                ("ProductVersion", str),
            ],
            (
                (indent, dataclasses.astuple(entry))
                for indent, entry in self._generator()
            ),
        )



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
import contextlib
import logging
from typing import List

from volatility3.framework import constants, exceptions, interfaces, objects, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import registry
from volatility3.plugins.windows import pslist
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)


class Envars(interfaces.plugins.PluginInterface):
    "Display process environment variables"

    _version = (1, 0, 0)
    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="silent",
                description="Suppress common and non-persistent variables",
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
        ]

    def _get_silent_vars(self) -> List[str]:
        """Enumerate persistent & common variables.

        This function collects the global (all users) and
        user-specific environment variables from the
        registry. Any variables in a process env block that
        does not exist in the persistent list was explicitly
        set with the SetEnvironmentVariable() API.
        """

        values = []

        for hive in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            hive_offsets=None,
        ):
            ## The global variables
            sys = None
            try:
                sys = hive.get_key(
                    "CurrentControlSet\\Control\\Session Manager\\Environment"
                )
            except (
                KeyError,
                registry.RegistryException,
            ):
                with contextlib.suppress(
                    KeyError,
                    registry.RegistryException,
                ):
                    sys = hive.get_key(
                        "ControlSet001\\Control\\Session Manager\\Environment"
                    )
            if sys:
                with contextlib.suppress(
                    KeyError,
                    registry.RegistryException,
                ):
                    for node in sys.get_values():
                        try:
                            value_node_name = node.get_name()
                            if value_node_name:
                                values.append(value_node_name)
                        except (
                            exceptions.InvalidAddressException,
                            registry.RegistryException,
                        ):
                            vollog.log(
                                constants.LOGLEVEL_VVV,
                                "Error while parsing global environment variables keys (some keys might be excluded)",
                            )
                            continue

            ntuser = None
            ## The user-specific variables
            with contextlib.suppress(
                KeyError,
                registry.RegistryException,
            ):
                ntuser = hive.get_key("Environment")
            if ntuser:
                with contextlib.suppress(
                    KeyError,
                    registry.RegistryException,
                ):
                    for node in ntuser.get_values():
                        try:
                            value_node_name = node.get_name()
                            if value_node_name:
                                values.append(value_node_name)
                        except (
                            exceptions.InvalidAddressException,
                            registry.RegistryException,
                        ):
                            vollog.log(
                                constants.LOGLEVEL_VVV,
                                "Error while parsing user environment variables keys (some keys might be excluded)",
                            )
                            continue

            ## The volatile user variables
            try:
                key = hive.get_key("Volatile Environment")
            except (
                KeyError,
                registry.RegistryException,
            ):
                continue
            try:
                for node in key.get_values():
                    try:
                        value_node_name = node.get_name()
                        if value_node_name:
                            values.append(value_node_name)
                    except (
                        exceptions.InvalidAddressException,
                        registry.RegistryException,
                    ):
                        vollog.log(
                            constants.LOGLEVEL_VVV,
                            "Error while parsing volatile environment variables keys (some keys might be excluded)",
                        )
                        continue
            except KeyError:
                continue

        ## These are variables set explicitly but are
        ## common enough to ignore safely.
        values.extend(
            [
                "ProgramFiles",
                "CommonProgramFiles",
                "SystemDrive",
                "SystemRoot",
                "ProgramData",
                "PUBLIC",
                "ALLUSERSPROFILE",
                "COMPUTERNAME",
                "SESSIONNAME",
                "USERNAME",
                "USERPROFILE",
                "PROMPT",
                "USERDOMAIN",
                "AppData",
                "CommonFiles",
                "CommonDesktop",
                "CommonProgramGroups",
                "CommonStartMenu",
                "CommonStartUp",
                "Cookies",
                "DesktopDirectory",
                "Favorites",
                "History",
                "NetHood",
                "PersonalDocuments",
                "RecycleBin",
                "StartMenu",
                "Templates",
                "AltStartup",
                "CommonFavorites",
                "ConnectionWizard",
                "DocAndSettingRoot",
                "InternetCache",
                "windir",
                "Path",
                "HOMEDRIVE",
                "PROCESSOR_ARCHITECTURE",
                "NUMBER_OF_PROCESSORS",
                "ProgramFiles(x86)",
                "CommonProgramFiles(x86)",
                "CommonProgramW6432",
                "PSModulePath",
                "PROCESSOR_IDENTIFIER",
                "FP_NO_HOST_CHECK",
                "LOCALAPPDATA",
                "TMP",
                "ProgramW6432",
            ]
        )

        return values

    def _generator(self, data):
        silent_vars = self._get_silent_vars() if self.config.get("SILENT") else []

        for task in data:
            for var, val in task.environment_variables():
                if var in silent_vars:
                    continue

                yield (
                    0,
                    (
                        int(task.UniqueProcessId),
                        str(objects.utility.array_to_string(task.ImageFileName)),
                        hex(task.get_peb().ProcessParameters.Environment.vol.offset),
                        str(var),
                        str(val),
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Block", str),
                ("Variable", str),
                ("Value", str),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=filter_func,
                )
            ),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from volatility3.framework import deprecation
from volatility3.plugins.windows.malware import svcdiff
from volatility3.plugins.windows import svcscan

vollog = logging.getLogger(__name__)


class SvcDiff(
    svcscan.SvcScan,
    deprecation.PluginRenameClass,
    replacement_class=svcdiff.SvcDiff,
    removal_date="2026-06-07",
):
    """Compares services found through list walking versus scanning to find rootkits (deprecated)."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._enumeration_method = self.service_diff

    _required_framework_version = (2, 4, 0)

    _version = (2, 0, 0)



# This file is Copyright 2022 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from volatility3.framework import interfaces, constants, configuration


class KernelModule(interfaces.automagic.AutomagicInterface):
    """Finds ModuleRequirements and ensures their layer, symbols and offsets"""

    priority = 100

    def __call__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        requirement: interfaces.configuration.RequirementInterface,
        progress_callback: constants.ProgressCallback = None,
    ) -> None:
        new_config_path = interfaces.configuration.path_join(
            config_path, requirement.name
        )
        if not isinstance(requirement, configuration.requirements.ModuleRequirement):
            # Check subrequirements
            for req in requirement.requirements:
                self(
                    context,
                    new_config_path,
                    requirement.requirements[req],
                    progress_callback,
                )
            return None
        if not requirement.unsatisfied(context, config_path):
            return None
        # The requirement is unfulfilled and is a ModuleRequirement

        context.config[interfaces.configuration.path_join(new_config_path, "class")] = (
            "volatility3.framework.contexts.Module"
        )

        for req in requirement.requirements:
            if (
                requirement.requirements[req].unsatisfied(context, new_config_path)
                and req != "offset"
            ):
                return None

        # We now just have the offset requirement, but the layer requirement has been fulfilled.
        # Unfortunately we don't know the layer name requirement's exact name

        for req in requirement.requirements:
            if isinstance(
                requirement.requirements[req],
                configuration.requirements.TranslationLayerRequirement,
            ):
                layer_kvo_config_path = interfaces.configuration.path_join(
                    new_config_path, req, "kernel_virtual_offset"
                )
                offset_config_path = interfaces.configuration.path_join(
                    new_config_path, "offset"
                )
                offset = context.config[layer_kvo_config_path]
                context.config[offset_config_path] = offset

        # Now construct the module based on the sub-requirements
        requirement.construct(context, config_path)



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3 import framework
from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.objects import utility

vollog = logging.getLogger(__name__)


class ModuleExtract(interfaces.plugins.PluginInterface):
    """Recreates an ELF file from a specific address in the kernel"""

    _version = (1, 0, 1)
    _required_framework_version = (2, 25, 0)

    framework.require_interface_version(*_required_framework_version)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.IntRequirement(
                name="base",
                description="Base virtual address to reconstruct an ELF file",
                optional=False,
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules_module_extract",
                version=(1, 0, 2),
                component=linux_utilities_modules.ModuleExtract,
            ),
        ]

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        base_address = self.config["base"]

        kernel_layer = self.context.layers[kernel.layer_name]

        if not kernel_layer.is_valid(base_address):
            vollog.error(
                f"Given base address ({base_address:#x}) is not valid in the kernel address space. Unable to extract file."
            )
            return

        module = kernel.object(object_type="module", offset=base_address, absolute=True)

        elf_data = linux_utilities_modules.ModuleExtract.extract_module(
            self.context, self.config["kernel"], module
        )
        if not elf_data:
            vollog.error(
                f"Unable to reconstruct the ELF for module struct at {base_address:#x}"
            )
            return

        module_name = utility.array_to_string(module.name)
        file_name = self.open.sanitize_filename(
            f"kernel_module.{module_name}.{base_address:#x}.elf"
        )

        with self.open(file_name) as file_handle:
            file_handle.write(elf_data)

        yield (
            0,
            (
                format_hints.Hex(base_address),
                len(elf_data),
                file_handle.preferred_filename,
            ),
        )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Base", format_hints.Hex),
                ("File Size", int),
                ("File output", str),
            ],
            self._generator(),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
import datetime
import dataclasses
from typing import List, Callable, Tuple, Iterable

from volatility3.framework import renderers, interfaces, constants
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.symbols import linux
from volatility3.plugins.linux import pslist
from volatility3.plugins import timeliner

vollog = logging.getLogger(__name__)


@dataclasses.dataclass
class FDUser:
    """FD user representation, featuring augmented information and formatted fields.
    This is the data the plugin will eventually display.
    """

    task_tgid: int
    task_tid: int
    task_comm: str
    fd_num: int
    full_path: str
    device: str = dataclasses.field(default=renderers.NotAvailableValue())
    inode_num: int = dataclasses.field(default=renderers.NotAvailableValue())
    inode_type: str = dataclasses.field(default=renderers.NotAvailableValue())
    file_mode: str = dataclasses.field(default=renderers.NotAvailableValue())
    change_time: datetime.datetime = dataclasses.field(
        default=renderers.NotAvailableValue()
    )
    modification_time: datetime.datetime = dataclasses.field(
        default=renderers.NotAvailableValue()
    )
    access_time: datetime.datetime = dataclasses.field(
        default=renderers.NotAvailableValue()
    )
    inode_size: int = dataclasses.field(default=renderers.NotAvailableValue())


@dataclasses.dataclass
class FDInternal:
    """FD internal representation containing only the core objects

    Fields:
        task: 'task_struct' object
        fd_fields: FD fields as obtained from LinuxUtilities.files_descriptors_for_process()
    """

    task: interfaces.objects.ObjectInterface
    fd_fields: Tuple[int, interfaces.objects.ObjectInterface, str]

    def to_user(self) -> FDUser:
        """Augment the FD information to be presented to the user

        Returns:
            An InodeUser dataclass
        """
        # Ensure all types are atomic immutable. Otherwise, astuple() will take a long
        # time doing a deepcopy of the Volatility objects.
        task_tgid = int(self.task.tgid)
        task_tid = int(self.task.pid)
        task_comm = utility.array_to_string(self.task.comm)
        fd_num, filp, full_path = self.fd_fields
        fd_num = int(fd_num)
        full_path = str(full_path)
        inode = filp.get_inode()
        if inode:
            superblock_ptr = inode.i_sb
            if superblock_ptr and superblock_ptr.is_readable():
                device = f"{superblock_ptr.major}:{superblock_ptr.minor}"
            else:
                device = renderers.NotAvailableValue()

            fd_user = FDUser(
                task_tgid=task_tgid,
                task_tid=task_tid,
                task_comm=task_comm,
                fd_num=fd_num,
                full_path=full_path,
                device=device,
                inode_num=int(inode.i_ino),
                inode_type=inode.get_inode_type() or renderers.UnparsableValue(),
                file_mode=inode.get_file_mode(),
                change_time=inode.get_change_time(),
                modification_time=inode.get_modification_time(),
                access_time=inode.get_access_time(),
                inode_size=int(inode.i_size),
            )
        else:
            # We use the dataclasses' default values
            fd_user = FDUser(
                task_tgid=task_tgid,
                task_tid=task_tid,
                task_comm=task_comm,
                fd_num=fd_num,
                full_path=full_path,
            )

        return fd_user


class Lsof(plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Lists open files for each processes."""

    _required_framework_version = (2, 0, 0)
    _version = (2, 1, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="files_only",
                description="Include only file descriptors of type file",
                optional=True,
                default=False,
            ),
        ]

    @classmethod
    def list_fds(
        cls,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
        filter_func: Callable[[int], bool] = lambda _: False,
        include_files_only: bool = False,
    ) -> Iterable[FDInternal]:
        """Enumerates open file descriptors in tasks

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate
            filter_func: A function which takes a process object and returns True if the process
                should be ignored/filtered

        Yields:
            A FDInternal object
        """
        linuxutils_symbol_table = None
        for task in pslist.PsList.list_tasks(
            context, vmlinux_module_name, filter_func, include_threads=True
        ):
            if linuxutils_symbol_table is None:
                if constants.BANG not in task.vol.type_name:
                    raise ValueError("Task is not part of a symbol table")
                linuxutils_symbol_table = task.vol.type_name.split(constants.BANG)[0]

            fd_generator = linux.LinuxUtilities.files_descriptors_for_process(
                context, linuxutils_symbol_table, task, files_only=include_files_only
            )

            for fd_fields in fd_generator:
                yield FDInternal(task=task, fd_fields=fd_fields)

    def _generator(self, pids, vmlinux_module_name, include_files_only):
        filter_func = pslist.PsList.create_pid_filter(pids)

        for fd_internal in self.list_fds(
            self.context,
            vmlinux_module_name,
            filter_func=filter_func,
            include_files_only=include_files_only,
        ):
            fd_user = fd_internal.to_user()
            yield (0, dataclasses.astuple(fd_user))

    def run(self):
        pids = self.config.get("pid", None)
        vmlinux_module_name = self.config["kernel"]
        include_files_only = self.config.get("files_only")

        tree_grid_args = [
            ("PID", int),
            ("TID", int),
            ("Process", str),
            ("FD", int),
            ("Path", str),
            ("Device", str),
            ("Inode", int),
            ("Type", str),
            ("Mode", str),
            ("Changed", datetime.datetime),
            ("Modified", datetime.datetime),
            ("Accessed", datetime.datetime),
            ("Size", int),
        ]
        return renderers.TreeGrid(
            tree_grid_args,
            self._generator(
                pids, vmlinux_module_name, include_files_only=include_files_only
            ),
        )

    def generate_timeline(self):
        pids = self.config.get("pid", None)
        vmlinux_module_name = self.config["kernel"]

        filter_func = pslist.PsList.create_pid_filter(pids)
        for fd_internal in self.list_fds(
            self.context, vmlinux_module_name, filter_func=filter_func
        ):
            fd_user = fd_internal.to_user()

            description = (
                f"Process {fd_user.task_comm} ({fd_user.task_tgid}/{fd_user.task_tid}) "
                f"Open '{fd_user.full_path}'"
            )

            yield description, timeliner.TimeLinerType.CHANGED, fd_user.change_time
            yield (
                description,
                timeliner.TimeLinerType.MODIFIED,
                fd_user.modification_time,
            )
            yield description, timeliner.TimeLinerType.ACCESSED, fd_user.access_time



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import re
from typing import List

from volatility3.framework import renderers, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.layers import scanners
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class VmaRegExScan(plugins.PluginInterface):
    """Scans all virtual memory areas for tasks using RegEx."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 2)

    MAXSIZE_DEFAULT = 128

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.StringRequirement(
                name="pattern", description="RegEx pattern", optional=False
            ),
            requirements.VersionRequirement(
                name="regex_scanner",
                component=scanners.RegExScanner,
                version=(1, 0, 0),
            ),
            requirements.IntRequirement(
                name="maxsize",
                description="Maximum size in bytes for displayed context",
                default=cls.MAXSIZE_DEFAULT,
                optional=True,
            ),
        ]

    def _generator(self, regex_pattern, tasks):
        regex_pattern = bytes(regex_pattern, "UTF-8")
        vollog.debug(f"RegEx Pattern: {regex_pattern}")

        for task in tasks:
            if not task.mm:
                continue
            name = utility.array_to_string(task.comm)

            # attempt to create a process layer for each task and skip those
            # that cannot (e.g. kernel threads)
            proc_layer_name = task.add_process_layer()
            if not proc_layer_name:
                continue

            # get the proc_layer object from the context
            proc_layer = self.context.layers[proc_layer_name]

            # get process sections for scanning
            sections = [
                (start, size) for (start, size) in task.get_process_memory_sections()
            ]

            for offset in proc_layer.scan(
                context=self.context,
                scanner=scanners.RegExScanner(regex_pattern),
                sections=sections,
                progress_callback=self._progress_callback,
            ):
                result_data = proc_layer.read(offset, self.MAXSIZE_DEFAULT, pad=True)

                # reapply the regex in order to extract just the match
                regex_result = re.match(regex_pattern, result_data)

                if regex_result:
                    # the match is within the results_data (e.g. it fits within MAXSIZE_DEFAULT)
                    # extract just the match itself
                    regex_match = regex_result.group(0)
                    text_result = str(regex_match, encoding="UTF-8", errors="replace")
                    bytes_result = regex_match
                else:
                    # the match is not with the results_data (e.g. it doesn't fit within MAXSIZE_DEFAULT)
                    text_result = str(result_data, encoding="UTF-8", errors="replace")
                    bytes_result = result_data

                user_pid = task.tgid
                yield (
                    0,
                    (
                        user_pid,
                        name,
                        format_hints.Hex(offset),
                        text_result,
                        bytes_result,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Offset", format_hints.Hex),
                ("Text", str),
                ("Hex", bytes),
            ],
            self._generator(
                self.config.get("pattern"),
                pslist.PsList.list_tasks(
                    self.context, self.config["kernel"], filter_func=filter_func
                ),
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Module to identify the Directory Table Base and architecture of windows
memory images.

This module contains a PageMapScanner that scans a physical layer to identify self-referential pointers.
All windows versions include a self-referential pointer in their Directory Table Base's top table, in order to
have a single offset that will allow manipulation of the page tables themselves.

In older windows version the self-referential pointer was at a specific fixed index within the table,
which was different for each architecture.  In very recent Windows versions, the self-referential pointer
index has been randomized, so a different heuristic must be used.  In these versions of windows it was found
that the physical offset for the DTB was always within the range of 0x1a0000 to 0x1b0000.  As such, a search
for any self-referential pointer within these pages gives a high probability of being an accurate DTB.

The self-referential indices for older versions of windows are listed below:

    +--------------+-------+
    | Architecture | Index |
    +==============+=======+
    | x86          | 0x300 |
    +--------------+-------+
    | PAE          | 0x3   |
    +--------------+-------+
    | x64          | 0x1ED |
    +--------------+-------+
"""

import logging
import struct
from typing import Generator, Iterable, List, Optional, Tuple, Type

from volatility3.framework import constants, interfaces, layers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import intel

vollog = logging.getLogger(__name__)


class DtbSelfReferential:
    """A generic DTB test which looks for a self-referential pointer at *any*
    index within the page."""

    def __init__(
        self,
        layer_type: Type[layers.intel.Intel],
        ptr_struct: str,
        mask: int,
        valid_range: Iterable[int],
        reserved_bits: int,
    ) -> None:
        self.layer_type = layer_type
        self.ptr_struct = ptr_struct
        self.ptr_size = struct.calcsize(ptr_struct)
        self.mask = mask
        self.page_size: int = layer_type.page_size
        self.valid_range = valid_range
        self.reserved_bits = reserved_bits

    def __call__(
        self, data: bytes, data_offset: int, page_offset: int
    ) -> Optional[Tuple[int, int]]:
        page = data[page_offset : page_offset + self.page_size]
        if not page:
            return None
        ref_pages = set()

        for ref in range(0, self.page_size, self.ptr_size):
            ptr_data = page[ref : ref + self.ptr_size]
            (ptr,) = struct.unpack(self.ptr_struct, ptr_data)
            # For both Intel-32e, bit 7 is reserved (more are reserved in PAE), so if that's ever set,
            # we can move on
            if (ptr & self.reserved_bits) and (ptr & 0x01):
                return None
            if ((ptr & self.mask) == (data_offset + page_offset)) and (
                data_offset + page_offset > 0
            ):
                # Pointer must be valid
                if ptr & 0x01:
                    ref_pages.add(ref)

        # The DTB is extremely unlikely to refer back to itself. so the number of reference should always be exactly 1
        if len(ref_pages) == 1:
            ref_page = ref_pages.pop()
            if (ref_page // self.ptr_size) in self.valid_range:
                return (data_offset + page_offset), ref_page
        return None


class DtbSelfRef32bit(DtbSelfReferential):
    def __init__(self):
        super().__init__(
            layer_type=layers.intel.WindowsIntel,
            ptr_struct="I",
            mask=0xFFFFF000,
            valid_range=[0x300],
            reserved_bits=0x0,
        )


class DtbSelfRef64bit(DtbSelfReferential):
    def __init__(self) -> None:
        super().__init__(
            layer_type=layers.intel.WindowsIntel32e,
            ptr_struct="Q",
            mask=0x3FFFFFFFFFF000,
            valid_range=range(0x100, 0x1FF),
            reserved_bits=0x80,
        )


class DtbSelfRef64bitOldWindows(DtbSelfReferential):
    def __init__(self) -> None:
        super().__init__(
            layer_type=layers.intel.WindowsIntel32e,
            ptr_struct="Q",
            mask=0x3FFFFFFFFFF000,
            valid_range=[0x1ED],
            reserved_bits=0x80,
        )


class DtbSelfRefPae(DtbSelfReferential):
    def __init__(self) -> None:
        super().__init__(
            layer_type=layers.intel.WindowsIntelPAE,
            ptr_struct="Q",
            valid_range=[0x3],
            mask=0x3FFFFFFFFFF000,
            reserved_bits=0x0,
        )

    @staticmethod
    def _and_bytes(abytes, bbytes):
        return bytes([a & b for a, b in zip(abytes[::-1], bbytes[::-1])][::-1])

    def __call__(
        self, data: bytes, data_offset: int, page_offset: int
    ) -> Optional[Tuple[int, int]]:
        dtb = super().__call__(data, data_offset, page_offset)
        if dtb:
            # Find the top page
            top_pae_page = dtb[0] - 0x4000
            # The top page should map to the next four pages after it
            # Build what we expect the page table to be
            expected_table = b"".join(
                [
                    struct.pack(self.ptr_struct, top_pae_page + (i * 0x1000))
                    for i in range(1, 5)
                ]
            )
            # Mask off the page bits of top level page map
            page_table_mask = b"\x00\xf0\xff\xff\xff\xff\xff\xff" * 4
            page_table = data[
                top_pae_page - data_offset : top_pae_page
                - data_offset
                + (4 * self.ptr_size)
            ]
            # Compare them
            anded_bytes = self._and_bytes(page_table, page_table_mask)
            if anded_bytes == expected_table:
                return top_pae_page, dtb[1]
            # Return None since the dtb value *isn't* None
            return None
        return dtb


class PageMapScanner(interfaces.layers.ScannerInterface):
    """Scans through all pages using DTB tests to determine a dtb offset and
    architecture."""

    overlap = 0x4000
    thread_safe = True
    tests = [DtbSelfRef64bit(), DtbSelfRefPae(), DtbSelfRef32bit()]
    """The default tests to run when searching for DTBs"""

    def __init__(self, tests: Optional[List[DtbSelfReferential]]) -> None:
        super().__init__()
        if tests:
            self.tests = tests

    def __call__(
        self, data: bytes, data_offset: int
    ) -> Generator[Tuple[DtbSelfReferential, int], None, None]:
        for page_offset in range(0, len(data), 0x1000):
            for test in self.tests:
                result = test(data, data_offset, page_offset)
                if result is not None:
                    yield (test, result[0])


class WindowsIntelStacker(interfaces.automagic.StackerLayerInterface):
    stack_order = 40
    exclusion_list = ["mac", "linux"]

    # Group these by region so we only run over the data once
    test_sets = [
        (
            "Detecting Self-referential pointer for recent windows",
            [DtbSelfRef64bit()],
            [(0x150000, 0x150000), (0x550000, 0x1A0000), (0x900000, 0x100000)],
        ),
        (
            "Older windows fixed location self-referential pointers",
            [DtbSelfRefPae(), DtbSelfRef32bit(), DtbSelfRef64bitOldWindows()],
            [(0x30000, 0x1000000)],
        ),
    ]

    @classmethod
    def stack(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[interfaces.layers.DataLayerInterface]:
        """Attempts to determine and stack an intel layer on a physical layer
        where possible.

        Where the DTB scan fails, it attempts a heuristic of checking
        for the DTB within a specific range. New versions of windows,
        with randomized self-referential pointers, appear to always load
        their dtb within a small specific range (`0x1a0000` and
        `0x1b0000`), so instead we scan for all self-referential
        pointers in that range, and ignore any that contain multiple
        self-references (since the DTB is very unlikely to point to
        itself more than once).
        """
        base_layer = context.layers[layer_name]
        if isinstance(base_layer, intel.Intel):
            return None
        if base_layer.metadata.get("os", None) not in ["Windows", "Unknown"]:
            return None
        layer = config_path = None

        # Check the metadata
        if base_layer.metadata.get("os", None) == "Windows" and base_layer.metadata.get(
            "page_map_offset"
        ):
            arch = base_layer.metadata.get("architecture", None)
            if arch not in ["Intel32", "Intel64"]:
                return None
            # Set the layer type
            layer_type: Type = intel.WindowsIntel
            if arch == "Intel64":
                layer_type = intel.WindowsIntel32e
            elif base_layer.metadata.get("pae", False):
                layer_type = intel.WindowsIntelPAE
            # Construct the layer
            new_layer_name = context.layers.free_layer_name("IntelLayer")
            config_path = interfaces.configuration.path_join(
                "IntelHelper", new_layer_name
            )
            context.config[
                interfaces.configuration.path_join(config_path, "memory_layer")
            ] = layer_name
            context.config[
                interfaces.configuration.path_join(config_path, "page_map_offset")
            ] = base_layer.metadata["page_map_offset"]
            layer = layer_type(
                context,
                config_path=config_path,
                name=new_layer_name,
                metadata={"os": "Windows"},
            )
            page_map_offset = context.config[
                interfaces.configuration.path_join(config_path, "page_map_offset")
            ]
            vollog.debug(f"DTB was given to us by base layer: {hex(page_map_offset)}")
            return layer

        # Self Referential finder
        for description, tests, sections in cls.test_sets:
            vollog.debug(description)
            # There is a very high chance that the DTB will live in these very narrow segments, assuming we couldn't find them previously
            hits = base_layer.scan(
                context,
                PageMapScanner(tests=tests),
                sections=sections,
                progress_callback=progress_callback,
            )

            # Flatten the generator
            def sort_by_tests(x):
                """Key used to sort by tests"""
                return tests.index(x[0]), x[1]

            def get_valid_page_table_pointers(page_table, ptr_size: int):
                """Yields valid pointers from a page table"""
                for index in range(0, len(page_table), ptr_size):
                    pointer = struct.unpack(
                        test.ptr_struct, page_table[index : index + ptr_size]
                    )[0]
                    # Make sure the pointer is valid, ignore large pages which would require more calculation
                    if pointer & 0x1 and not pointer & 0x80:
                        yield pointer

            def get_max_pointer(page_table, test, ptr_size: int):
                """Determines a pointer from a page_table"""
                max_ptr = 0
                for pointer in get_valid_page_table_pointers(page_table, ptr_size):
                    max_ptr = max(
                        max_ptr,
                        (pointer ^ (pointer & 0xFFF)) % test.layer_type.maximum_address,
                    )
                return max_ptr

            def page_table_is_dummy(page_table, ptr_size: int):
                """Verify that a page table has at least 12 valid pointers"""
                valid_pointers = 0
                for _ in get_valid_page_table_pointers(page_table, ptr_size):
                    valid_pointers += 1
                    # 10 is an arbitrary constant
                    if valid_pointers >= 10:
                        # Do not consume the entire generator to enhance performance
                        return False
                vollog.debug(f"Found {valid_pointers} valid pointers")
                return True

            hits = sorted(list(hits), key=sort_by_tests)

            vollog.debug(f"WindowsIntelStacker hits: {hits}")

            for test, page_map_offset in hits:
                # Turn the page tables into integers and find the largest one
                page_table = base_layer.read(page_map_offset, 0x1000)
                ptr_size = struct.calcsize(test.ptr_struct)
                # Modern windows can have a dummy page table with only about 2 entries, so sanity check
                if page_table_is_dummy(page_table, ptr_size):
                    vollog.debug(
                        f"DTB {page_map_offset:x} contains less than 12 valid pointers, ignoring"
                    )
                    continue

                max_pointer = get_max_pointer(page_table, test, ptr_size)

                if max_pointer <= base_layer.maximum_address:
                    vollog.debug(
                        f"{test.__class__.__name__} test succeeded at {hex(page_map_offset)}"
                    )
                    new_layer_name = context.layers.free_layer_name("IntelLayer")
                    config_path = interfaces.configuration.path_join(
                        "IntelHelper", new_layer_name
                    )
                    context.config[
                        interfaces.configuration.path_join(config_path, "memory_layer")
                    ] = layer_name
                    context.config[
                        interfaces.configuration.path_join(
                            config_path, "page_map_offset"
                        )
                    ] = page_map_offset
                    layer = test.layer_type(
                        context,
                        config_path=config_path,
                        name=new_layer_name,
                        metadata={"os": "Windows"},
                    )
                    break
                else:
                    vollog.debug(
                        f"Max pointer for hit with test {test.__class__.__name__} not met: {hex(max_pointer)} > {hex(base_layer.maximum_address)}"
                    )
            if layer is not None and config_path:
                break

        if layer is not None and config_path:
            vollog.debug(
                "DTB was found at: 0x{:0x}".format(
                    context.config[
                        interfaces.configuration.path_join(
                            config_path, "page_map_offset"
                        )
                    ]
                )
            )
        return layer


class WinSwapLayers(interfaces.automagic.AutomagicInterface):
    """Class to read swap_layers filenames from single-swap-layers, create the
    layers and populate the single-layers swap_layers."""

    exclusion_list = ["linux", "mac"]

    def __call__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        requirement: interfaces.configuration.RequirementInterface,
        progress_callback: constants.ProgressCallback = None,
    ) -> None:
        """Finds translation layers that can have swap layers added."""

        path_join = interfaces.configuration.path_join
        self._translation_requirement = self.find_requirements(
            context,
            config_path,
            requirement,
            requirements.TranslationLayerRequirement,
            shortcut=False,
        )
        for trans_sub_config, trans_req in self._translation_requirement:
            if not isinstance(trans_req, requirements.TranslationLayerRequirement):
                # We need this so the type-checker knows we're a TranslationLayerRequirement
                continue
            swap_sub_config, swap_req = self.find_swap_requirement(
                trans_sub_config, trans_req
            )

            counter = 0
            swap_config = interfaces.configuration.parent_path(swap_sub_config)

            if swap_req and swap_req.unsatisfied(context, swap_config):
                # See if any of them need constructing

                for swap_location in self.config.get("single_swap_locations", []):
                    # Setup config locations/paths
                    current_layer_name = swap_req.name + str(counter)
                    current_layer_path = path_join(swap_sub_config, current_layer_name)
                    layer_loc_path = path_join(current_layer_path, "location")
                    layer_class_path = path_join(current_layer_path, "class")
                    counter += 1

                    # Fill in the config
                    if swap_location:
                        context.config[current_layer_path] = current_layer_name
                        try:
                            context.config[layer_loc_path] = (
                                requirements.URIRequirement.location_from_file(
                                    swap_location
                                )
                            )
                        except ValueError:
                            vollog.warning(
                                f"Volatility swap_location {swap_location} could not be validated - swap layer disabled"
                            )
                            continue
                        context.config[layer_class_path] = (
                            "volatility3.framework.layers.physical.FileLayer"
                        )

                    # Add the requirement
                    new_req = requirements.TranslationLayerRequirement(
                        name=current_layer_name,
                        description="Swap Layer",
                        optional=False,
                    )
                    swap_req.add_requirement(new_req)

                context.config[path_join(swap_sub_config, "number_of_elements")] = (
                    counter
                )
                context.config[swap_sub_config] = True

                swap_req.construct(context, swap_config)

    @staticmethod
    def find_swap_requirement(
        config: str, requirement: requirements.TranslationLayerRequirement
    ) -> Tuple[str, Optional[requirements.LayerListRequirement]]:
        """Takes a Translation layer and returns its swap_layer requirement."""
        swap_req = None
        for req_name in requirement.requirements:
            req = requirement.requirements[req_name]
            if (
                isinstance(req, requirements.LayerListRequirement)
                and req.name == "swap_layers"
            ):
                swap_req = req
                continue

        swap_config = interfaces.configuration.path_join(config, "swap_layers")
        return swap_config, swap_req

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        """Returns the requirements of this plugin."""
        return [
            requirements.ListRequirement(
                name="single_swap_locations",
                element_type=str,
                min_elements=0,
                max_elements=16,
                description="Specifies a list of swap layer URIs for use with single-location",
                optional=True,
            )
        ]



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Iterable, Tuple

from volatility3.framework import interfaces, renderers, constants, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.objects import utility
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.linux.extensions import elf
from volatility3.plugins.linux import pslist


vollog = logging.getLogger(__name__)


class LibraryList(interfaces.plugins.PluginInterface):
    """Enumerate libraries loaded into processes"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 2)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pids",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
        ]

    def _get_libdl_libraries(
        self, proc_layer_name: str, vma_start: int
    ) -> interfaces.objects.ObjectInterface:
        """Get the ELF link map objects for the given VMA address

        Args:
            proc_layer_name (str): Name of the process layer
            vma_start (int): VMA start address

        Yields:
            ELF link map objects for the given VMA address
        """
        elf_table_name = intermed.IntermediateSymbolTable.create(
            self.context,
            self.config_path,
            "linux",
            "elf",
            class_types=elf.class_types,
        )
        elf_object = self.context.object(
            elf_table_name + constants.BANG + "Elf",
            offset=vma_start,
            layer_name=proc_layer_name,
        )

        if not elf_object or not elf_object.is_valid():
            return None

        kernel = self.context.modules[self.config["kernel"]]

        try:
            for link_map in elf_object.get_link_maps(kernel.symbol_table_name):
                if link_map.l_addr and link_map.l_name:
                    yield link_map
        except exceptions.InvalidAddressException:
            # Protection against memory smear in this VMA
            pass

    def _get_libdl_maps(
        self, task: interfaces.objects.ObjectInterface, proc_layer_name: str
    ) -> interfaces.objects.ObjectInterface:
        """Get the ELF link maps objects for a task

        Args:
            task (task_struct): A reference task
            proc_layer_name (str): Name of the process layer

        Yields:
            ELF link map objects
        """

        link_map_seen = set()
        for vma in task.mm.get_vma_iter():
            for link_map in self._get_libdl_libraries(proc_layer_name, vma.vm_start):
                if link_map.l_addr in link_map_seen:
                    continue

                yield link_map
                link_map_seen.add(link_map.l_addr)

    def _get_task_libraries(
        self, task: interfaces.objects.ObjectInterface
    ) -> Tuple[int, str]:
        """Get the task libraries from the ELF headers found within the memory maps

        Args:
            task (task_struct): The reference task

        Yields:
            Tuples with a ELF link map address and name
        """
        proc_layer_name = task.add_process_layer()
        if not proc_layer_name:
            return

        for elf_link_map in self._get_libdl_maps(task, proc_layer_name):
            name = elf_link_map.get_name()
            if not name:
                continue
            yield elf_link_map.l_addr, name

    def _get_tasks_libraries(
        self,
        tasks: Iterable[interfaces.objects.ObjectInterface],
    ) -> Iterable[Tuple[str, int, int, str]]:
        """Get the task libraries from the ELF headers found within the memory maps for
        all the tasks.

        Args:
            tasks: An iterable of tasks

        Yields:
            Tuples with a task name, task tgid, an ELF link map address and name
        """
        for task in tasks:
            task_name = utility.array_to_string(task.comm)
            for linkmap_addr, linkmap_name in self._get_task_libraries(task):
                yield task_name, task.tgid, linkmap_addr, linkmap_name

    def _format_fields(self, fields):
        task_name, task_pid, addr, name = fields
        return task_name, task_pid, format_hints.Hex(addr), name

    def _generator(
        self, tasks: Iterable[interfaces.objects.ObjectInterface]
    ) -> Iterable[Tuple[int, Tuple]]:
        for fields in self._get_tasks_libraries(tasks):
            yield 0, self._format_fields(fields)

    def run(self):
        pids = self.config.get("pids")
        pid_filter = pslist.PsList.create_pid_filter(pids)
        tasks = pslist.PsList.list_tasks(
            self.context, self.config["kernel"], filter_func=pid_filter
        )

        headers = [
            ("Name", str),
            ("Pid", int),
            ("LoadAddress", format_hints.Hex),
            ("Path", str),
        ]

        return renderers.TreeGrid(headers, self._generator(tasks))



# This file is Copyright 2023 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from dataclasses import dataclass, astuple, fields
from typing import Iterable, List, Tuple

from volatility3.framework import interfaces, renderers, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.symbols.linux import extensions
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


@dataclass
class TaskData:
    """Stores basic information about a task"""

    comm: str
    pid: int
    tgid: int
    ppid: int
    euid: int


@dataclass
class CapabilitiesData:
    """Stores each set of capabilities for a task"""

    cap_inheritable: interfaces.objects.ObjectInterface
    cap_permitted: interfaces.objects.ObjectInterface
    cap_effective: interfaces.objects.ObjectInterface
    cap_bset: interfaces.objects.ObjectInterface
    cap_ambient: interfaces.objects.ObjectInterface

    def astuple(self) -> Tuple:
        """Returns a shallow copy of the capability sets in a tuple.

        Otherwise, when dataclasses.astuple() performs a deep-copy recursion on
        ObjectInterface will take a substantial amount of time.
        """
        return tuple(getattr(self, field.name) for field in fields(self))


class Capabilities(plugins.PluginInterface):
    """Lists process capabilities"""

    _required_framework_version = (2, 13, 0)
    _version = (1, 1, 1)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pids",
                description="Filter on specific process IDs.",
                element_type=int,
                optional=True,
            ),
        ]

    def _check_capabilities_support(
        self,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
    ):
        """Checks that the framework supports at least as much capabilities as
        the kernel being analysed. Otherwise, it shows a warning for the
        developers.
        """

        vmlinux = context.modules[vmlinux_module_name]

        try:
            kernel_cap_last_cap = vmlinux.object_from_symbol(symbol_name="cap_last_cap")
        except exceptions.SymbolError:
            # It should be a kernel < 3.2 See 73efc0394e148d0e15583e13712637831f926720
            return None

        vol2_last_cap = extensions.kernel_cap_struct.get_last_cap_value()
        if kernel_cap_last_cap > vol2_last_cap:
            vollog.warning(
                "Developers: The supported Linux capabilities of this plugin are outdated for this kernel"
            )

    @staticmethod
    def _decode_cap(cap: interfaces.objects.ObjectInterface) -> str:
        """Returns a textual representation of the capability set.
        The format is a comma-separated list of capabilitites. In order to
        summarize the output and if all the capabilities are enabled, instead of
        the individual capabilities, the special name "all" will be shown.

        Args:
            cap: Kernel capability object. Usually a 'kernel_cap_struct' struct

        Returns:
            str: A string with a comma separated list of decoded capabilities
        """
        if isinstance(cap, renderers.NotAvailableValue):
            return cap

        cap_value = cap.get_capabilities()
        if not cap_value:
            return ""

        if cap_value == cap.get_kernel_cap_full():
            return "all"

        return ", ".join(cap.enumerate_capabilities())

    @classmethod
    def get_task_capabilities(
        cls, task: interfaces.objects.ObjectInterface
    ) -> Tuple[TaskData, CapabilitiesData]:
        """Returns a tuple with the task basic information along with its capabilities

        Args:
            task: A task object from where to get the fields.

        Returns:
            A tuple with the task basic information and its capabilities
        """
        task_data = TaskData(
            comm=utility.array_to_string(task.comm),
            pid=int(task.pid),
            tgid=int(task.tgid),
            ppid=int(task.get_parent_pid()),
            euid=int(task.cred.euid),
        )

        task_cred = task.real_cred
        capabilities_data = CapabilitiesData(
            cap_inheritable=task_cred.cap_inheritable,
            cap_permitted=task_cred.cap_permitted,
            cap_effective=task_cred.cap_effective,
            cap_bset=task_cred.cap_bset,
            cap_ambient=renderers.NotAvailableValue(),
        )

        # Ambient capabilities were added in kernels 4.3.6
        if task_cred.has_member("cap_ambient"):
            capabilities_data.cap_ambient = task_cred.cap_ambient

        return task_data, capabilities_data

    @classmethod
    def get_tasks_capabilities(
        cls, tasks: List[interfaces.objects.ObjectInterface]
    ) -> Iterable[Tuple[TaskData, CapabilitiesData]]:
        """Yields a tuple for each task containing the task's basic information along with its capabilities

        Args:
            tasks: An iterable with the tasks to process.

        Yields:
            A tuple for each task containing the task's basic information and its capabilities
        """
        for task in tasks:
            yield cls.get_task_capabilities(task)

    def _generator(
        self, tasks: Iterable[interfaces.objects.ObjectInterface]
    ) -> Iterable[Tuple[int, Tuple]]:
        for task_fields, capabilities_fields in self.get_tasks_capabilities(tasks):
            task_fields = astuple(task_fields)

            capabilities_text = tuple(
                self._decode_cap(cap) for cap in capabilities_fields.astuple()
            )

            yield 0, task_fields + capabilities_text

    def run(self):
        self._check_capabilities_support(self.context, self.config["kernel"])

        pids = self.config.get("pids")
        pid_filter = pslist.PsList.create_pid_filter(pids)
        tasks = pslist.PsList.list_tasks(
            self.context, self.config["kernel"], filter_func=pid_filter
        )

        columns = [
            ("Name", str),
            ("Tid", int),
            ("Pid", int),
            ("PPid", int),
            ("EUID", int),
            ("cap_inheritable", str),
            ("cap_permitted", str),
            ("cap_effective", str),
            ("cap_bounding", str),
            ("cap_ambient", str),
        ]

        return renderers.TreeGrid(columns, self._generator(tasks))



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A module containing a plugin that checks the system call table for hooks."""

import contextlib
import logging
from typing import List

from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.renderers import format_hints

vollog = logging.getLogger(__name__)

try:
    import capstone

    has_capstone = True
except ImportError:
    has_capstone = False


class Check_syscall(plugins.PluginInterface):
    """Check system call table for hooks."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
        ]

    def _get_table_size_next_symbol(self, table_addr, ptr_sz, vmlinux):
        """Returns the size of the table based on the next symbol."""
        ret = 0

        symbol_list = []
        for sn in vmlinux.symbols:
            with contextlib.suppress(exceptions.SymbolError):
                # When requesting the symbol from the module, a full resolve is performed
                symbol_list.append((vmlinux.get_symbol(sn).address, sn))
        sorted_symbols = sorted(symbol_list)

        sym_address = 0

        for tmp_sym_address, sym_name in sorted_symbols:
            if tmp_sym_address > table_addr:
                sym_address = tmp_sym_address
                break

        if sym_address > 0:
            ret = int((sym_address - table_addr) / ptr_sz)

        return ret

    def _get_table_size_meta(self, vmlinux):
        """returns the number of symbols that start with __syscall_meta__ this
        is a fast way to determine the number of system calls, but not the most
        accurate."""

        return len(
            [
                sym
                for sym in self.context.symbol_space[vmlinux.symbol_table_name].symbols
                if sym.startswith("__syscall_meta__")
            ]
        )

    def _get_table_info_other(self, table_addr, ptr_sz, vmlinux):
        table_size_meta = self._get_table_size_meta(vmlinux)
        table_size_syms = self._get_table_size_next_symbol(table_addr, ptr_sz, vmlinux)

        sizes = [size for size in [table_size_meta, table_size_syms] if size > 0]

        table_size = min(sizes)

        return table_size

    def _get_table_info_disassembly(self, ptr_sz, vmlinux) -> int:
        """Find the size of the system call table by disassembling functions
        that immediately reference it in their first instruction This is in the
        form 'cmp reg,NR_syscalls'."""
        table_size = 0

        if not has_capstone:
            return table_size

        if ptr_sz == 4:
            syscall_entry_func = "sysenter_do_call"
            mode = capstone.CS_MODE_32
        else:
            syscall_entry_func = "system_call_fastpath"
            mode = capstone.CS_MODE_64

        md = capstone.Cs(capstone.CS_ARCH_X86, mode)

        try:
            func_addr = vmlinux.get_symbol(syscall_entry_func).address
        except exceptions.SymbolError:
            # if we can't find the disassemble function then bail and rely on a different method
            return 0

        vmlinux = self.context.modules[self.config["kernel"]]
        vmlinux_layer = self.context.layers[vmlinux.layer_name]
        try:
            data = vmlinux_layer.read(func_addr, 6)
        except exceptions.InvalidAddressException:
            return 0

        for _address, _size, mnemonic, op_str in md.disasm_lite(data, func_addr):
            if mnemonic == "CMP":
                table_size = int(op_str.split(",")[1].strip()) & 0xFFFF
                break

        return table_size

    def _get_table_info(self, vmlinux, table_name, ptr_sz):
        table_sym = vmlinux.get_symbol(table_name)

        table_size = self._get_table_info_disassembly(ptr_sz, vmlinux)

        if table_size == 0:
            table_size = self._get_table_info_other(table_sym.address, ptr_sz, vmlinux)

            if table_size == 0:
                vollog.error("Unable to get system call table size")
                return 0, 0

        return table_sym.address, table_size

    # TODO - add finding and parsing unistd.h once cached file enumeration is added
    def _generator(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        ptr_sz = vmlinux.get_type("pointer").size
        if ptr_sz == 4:
            table_name = "32bit"
        else:
            table_name = "64bit"

        try:
            table_info = self._get_table_info(vmlinux, "sys_call_table", ptr_sz)
        except exceptions.SymbolError:
            vollog.error("Unable to find the system call table. Exiting.")
            return None

        tables = [(table_name, table_info)]

        # this table is only present on 64 bit systems with 32 bit emulation
        # enabled in order to support 32 bit programs and libraries
        # if the symbol isn't there then the support isn't in the kernel and so we skip it
        try:
            ia32_symbol = vmlinux.get_symbol("ia32_sys_call_table")
        except exceptions.SymbolError:
            ia32_symbol = None

        if ia32_symbol is not None:
            ia32_info = self._get_table_info(vmlinux, "ia32_sys_call_table", ptr_sz)
            tables.append(("32bit", ia32_info))

        for table_name, (tableaddr, tblsz) in tables:
            table = vmlinux.object(
                object_type="array",
                subtype=vmlinux.get_type("pointer"),
                offset=tableaddr,
                count=tblsz,
            )

            for i in range(len(table)):
                try:
                    call_addr = table[i]
                except exceptions.InvalidAddressException:
                    vollog.debug(f"Failed to get system call table entry at index {i}")
                    continue

                symbols = list(vmlinux.get_symbols_by_absolute_location(call_addr))

                if len(symbols) > 0:
                    sym_name = (
                        str(symbols[0].split(constants.BANG)[1])
                        if constants.BANG in symbols[0]
                        else str(symbols[0])
                    )
                else:
                    sym_name = "UNKNOWN"

                yield (
                    0,
                    (
                        format_hints.Hex(tableaddr),
                        table_name,
                        i,
                        format_hints.Hex(call_addr),
                        sym_name,
                    ),
                )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Table Address", format_hints.Hex),
                ("Table Name", str),
                ("Index", int),
                ("Handler Address", format_hints.Hex),
                ("Handler Symbol", str),
            ],
            self._generator(),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List, Iterable

from volatility3.framework import interfaces, renderers, exceptions
from volatility3.framework.objects import utility
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import windowstations

vollog = logging.getLogger(__name__)


class Windows(interfaces.plugins.PluginInterface):
    """Enumerates the Windows of Desktop instances"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="windowstations",
                component=windowstations.WindowStations,
                version=(1, 0, 0),
            ),
        ]

    @classmethod
    def list_windows(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """
        Enumerates the desktops of each window station
        For each found, enumerates its windows within the desktop
        """
        kernel = context.modules[kernel_module_name]

        for (
            winsta,
            station_name,
            session_id,
        ) in windowstations.WindowStations.scan_window_stations(
            context, config_path, kernel_module_name
        ):
            # for each window station, walk its list of desktops
            for desktop, desktop_name in winsta.desktops(kernel.symbol_table_name):
                try:
                    top_window = desktop.pDeskInfo.spwnd
                except exceptions.InvalidAddressException:
                    vollog.debug(
                        f"Desktop with name {desktop_name} in window station {station_name} has a broken window pointer."
                    )
                    continue

                for window, window_name in desktop.windows(top_window):
                    yield station_name, desktop_name, window, window_name

    def _generator(self):
        kernel_name = self.config["kernel"]

        # call the implementation for finding windows and gather attributes
        for station_name, desktop_name, window, window_name in self.list_windows(
            self.context, self.config_path, kernel_name
        ):
            # We need a valid process and session id for the window to display it
            process = window.get_process()
            process_name = None
            if process:
                try:
                    process_name = utility.array_to_string(process.ImageFileName)
                    process_pid = process.UniqueProcessId
                except exceptions.InvalidAddressException:
                    vollog.debug(
                        f"Unable to read name and pid of the process for window {window.vol.offset:#x}"
                    )

            if process_name is None:
                vollog.debug(
                    f"Invalid process reference for the process hosting window {window.vol.offset:#x}"
                )
                continue

            sess_id = window.get_session_id()
            if sess_id is None:
                vollog.debug(
                    f"Unable to read session id of the process for window {window.vol.offset:#x} in process {process_name}"
                )
                continue

            # procedures can be empty, but if set, should be a valid pointer
            window_proc = window.get_window_procedure()
            if window_proc is None:
                window_proc = renderers.NotAvailableValue()
            elif window_proc == 0 or window_proc > 0x1000:
                window_proc = format_hints.Hex(window_proc)
            else:
                vollog.debug(
                    f"Invalid window procedure {window_proc} for the window {window.vol.offset:#x}"
                )
                continue

            yield (
                0,
                (
                    format_hints.Hex(window.vol.offset),
                    station_name,
                    sess_id,
                    desktop_name,
                    window_name or renderers.NotAvailableValue(),
                    window_proc,
                    process_name,
                    process_pid,
                ),
            )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Station", str),
                ("Session", int),
                ("Desktop", str),
                ("Window", str),
                ("Procedure", format_hints.Hex),
                ("Process", str),
                ("PID", int),
            ],
            self._generator(),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import linux
from volatility3.framework.constants import architectures
from volatility3.framework.objects import utility
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class Kthreads(plugins.PluginInterface):
    """Enumerates kthread functions"""

    _required_framework_version = (2, 11, 0)
    _version = (1, 0, 3)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules",
                component=linux_utilities_modules.Modules,
                version=(3, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherers",
                component=linux_utilities_modules.ModuleGatherers,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 1, 0)
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
        ]

    def _generator(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        kthread_type = vmlinux.get_type("kthread")

        if not kthread_type.has_member("threadfn"):
            raise exceptions.VolatilityException(
                "Unsupported kthread implementation. This plugin only works with kernels >= 5.8"
            )

        known_modules = linux_utilities_modules.Modules.run_modules_scanners(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            caller_wanted_gatherers=linux_utilities_modules.ModuleGatherers.all_gatherers_identifier,
        )

        for task in pslist.PsList.list_tasks(
            self.context, vmlinux.name, include_threads=True
        ):
            if not task.is_kernel_thread:
                continue

            if task.has_member("worker_private"):
                # kernels >= 5.17 e32cf5dfbe227b355776948b2c9b5691b84d1cbd
                kthread_base_pointer = task.worker_private
            else:
                # 5.8 <= kernels < 5.17 in 52782c92ac85c4e393eb4a903a62e6c24afa633f threadfn
                # was added to struct kthread. task.set_child_tid is safe on those versions.
                kthread_base_pointer = task.set_child_tid

            if not kthread_base_pointer.is_readable():
                continue

            kthread = kthread_base_pointer.dereference().cast("kthread")
            threadfn = kthread.threadfn
            if not (threadfn and threadfn.is_readable()):
                continue

            thread_name = utility.array_to_string(task.comm)

            # kernels >= 5.17 in d6986ce24fc00b0638bd29efe8fb7ba7619ed2aa full_name was added to kthread
            if kthread.has_member("full_name"):
                try:
                    thread_name = utility.pointer_to_string(
                        kthread.full_name, count=255
                    )
                except exceptions.InvalidAddressException:
                    vollog.debug(
                        f"full_name pointer for thread at {kthread.vol.offset:#x} is paged out."
                    )

            module_info, symbol_name = (
                linux_utilities_modules.Modules.module_lookup_by_address(
                    self.context, vmlinux.name, known_modules, threadfn
                )
            )

            if module_info:
                module_name = module_info.name
            else:
                module_name = renderers.NotAvailableValue()

            fields = [
                task.pid,
                thread_name,
                format_hints.Hex(threadfn),
                module_name,
                symbol_name or renderers.NotAvailableValue(),
            ]
            yield 0, fields

    def run(self):
        return renderers.TreeGrid(
            [
                ("TID", int),
                ("Thread Name", str),
                ("Handler Address", format_hints.Hex),
                ("Module", str),
                ("Symbol", str),
            ],
            self._generator(),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import List, Iterable

from volatility3.framework import renderers, interfaces, constants
from volatility3.framework.symbols import linux
from volatility3.framework.renderers import format_hints
from volatility3.framework.interfaces import plugins
from volatility3.framework.configuration import requirements
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class PIDHashTable(plugins.PluginInterface):
    """Enumerates processes through the PID hash table"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 3)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 1, 0)
            ),
            requirements.BooleanRequirement(
                name="decorate_comm",
                description="Show `user threads` comm in curly brackets, and `kernel threads` comm in square brackets",
                optional=True,
                default=False,
            ),
        ]

    def _is_valid_task(self, task) -> bool:
        return bool(task and task.pid > 0 and task.parent.is_readable())

    def _get_pidtype_pid(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        # The pid_type enumeration is present since 2.5.37, just in case
        pid_type_enum = vmlinux.get_enumeration("pid_type")
        if not pid_type_enum:
            vollog.error("Cannot find pid_type enum. Unsupported kernel")
            return None

        pidtype_pid = pid_type_enum.choices.get("PIDTYPE_PID")
        if pidtype_pid is None:
            vollog.error("Cannot find PIDTYPE_PID. Unsupported kernel")
            return None

        # Typically PIDTYPE_PID = 0
        return pidtype_pid

    def _get_pidhash_array(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        pidhash_shift = vmlinux.object_from_symbol("pidhash_shift")
        pidhash_size = 1 << pidhash_shift

        array_type_name = vmlinux.symbol_table_name + constants.BANG + "array"

        pidhash_ptr = vmlinux.object_from_symbol("pid_hash")
        # pidhash is an array of hlist_heads
        pidhash = self._context.object(
            array_type_name,
            offset=pidhash_ptr,
            subtype=vmlinux.get_type("hlist_head"),
            count=pidhash_size,
            layer_name=vmlinux.layer_name,
        )

        return pidhash

    def _walk_upid(self, seen_upids, upid):
        vmlinux = self.context.modules[self.config["kernel"]]
        vmlinux_layer = self.context.layers[vmlinux.layer_name]

        while upid and vmlinux_layer.is_valid(upid.vol.offset):
            if upid.vol.offset in seen_upids:
                break
            seen_upids.add(upid.vol.offset)

            pid_chain = upid.pid_chain
            if not (pid_chain.next and pid_chain.next.is_readable()):
                break

            upid = linux.LinuxUtilities.container_of(
                pid_chain.next, "upid", "pid_chain", vmlinux
            )

    def _get_upids(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        # 2.6.24 <= kernels < 4.15
        pidhash = self._get_pidhash_array()

        seen_upids = set()
        for hlist in pidhash:
            # each entry in the hlist is a upid which is wrapped in a pid
            ent = hlist.first

            while ent and ent.is_readable():
                # upid->pid_chain exists 2.6.24 <= kernel < 4.15
                upid = linux.LinuxUtilities.container_of(
                    ent.vol.offset, "upid", "pid_chain", vmlinux
                )

                if upid.vol.offset in seen_upids:
                    break

                self._walk_upid(seen_upids, upid)

                ent = ent.next

        return seen_upids

    def _pid_hash_implementation(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        # 2.6.24 <= kernels < 4.15
        task_pids_off = vmlinux.get_type("task_struct").relative_child_offset("pids")
        pidtype_pid = self._get_pidtype_pid()

        for upid in self._get_upids():
            pid = linux.LinuxUtilities.container_of(upid, "pid", "numbers", vmlinux)
            if not pid:
                continue

            pid_tasks_0 = pid.tasks[pidtype_pid].first
            if not (pid_tasks_0 and pid_tasks_0.is_readable()):
                continue

            task = vmlinux.object(
                "task_struct", offset=pid_tasks_0 - task_pids_off, absolute=True
            )
            if self._is_valid_task(task):
                yield task

    def _task_for_radix_pid_node(self, nodep):
        vmlinux = self.context.modules[self.config["kernel"]]

        # kernels >= 4.15
        pid = vmlinux.object("pid", offset=nodep, absolute=True)
        pidtype_pid = self._get_pidtype_pid()

        pid_tasks_0 = pid.tasks[pidtype_pid].first
        if not (pid_tasks_0 and pid_tasks_0.is_readable()):
            return None

        task_struct_type = vmlinux.get_type("task_struct")
        if task_struct_type.has_member("pids"):
            member = "pids"
        elif task_struct_type.has_member("pid_links"):
            member = "pid_links"
        else:
            return None

        task_pids_off = task_struct_type.relative_child_offset(member)
        task = vmlinux.object(
            "task_struct", offset=pid_tasks_0 - task_pids_off, absolute=True
        )
        return task

    def _pid_namespace_idr(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        # kernels >= 4.15
        ns_addr = vmlinux.get_symbol("init_pid_ns").address
        ns = vmlinux.object("pid_namespace", offset=ns_addr)

        for page_addr in ns.idr.get_entries():
            task = self._task_for_radix_pid_node(page_addr)
            if self._is_valid_task(task):
                yield task

    def _determine_pid_func(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        pid_hash = vmlinux.has_symbol("pid_hash") and vmlinux.has_symbol(
            "pidhash_shift"
        )  # 2.5.55 <= kernels < 4.15

        has_pid_numbers = vmlinux.has_type("pid") and vmlinux.get_type(
            "pid"
        ).has_member("numbers")  # kernels >= 2.6.24

        has_pid_chain = vmlinux.has_type("upid") and vmlinux.get_type(
            "upid"
        ).has_member("pid_chain")  # 2.6.24 <= kernels < 4.15

        # kernels >= 4.15
        pid_idr = vmlinux.has_type("pid_namespace") and vmlinux.get_type(
            "pid_namespace"
        ).has_member("idr")

        if pid_idr:
            # kernels >= 4.15
            return self._pid_namespace_idr
        elif pid_hash and has_pid_numbers and has_pid_numbers and has_pid_chain:
            # 2.6.24 <= kernels < 4.15
            return self._pid_hash_implementation

        return None

    def get_tasks(self) -> Iterable[interfaces.objects.ObjectInterface]:
        """Enumerates processes through the PID hash table

        Yields:
            task_struct objects
        """
        pid_func = self._determine_pid_func()
        if not pid_func:
            vollog.error("Cannot determine which PID hash table this kernel is using")
            return

        yield from sorted(pid_func(), key=lambda t: (t.tgid, t.pid))

    def _generator(self, decorate_comm: bool = False):
        for task in self.get_tasks():
            task_fields = pslist.PsList.get_task_fields(task, decorate_comm)
            fields = (
                format_hints.Hex(task_fields.offset),
                task_fields.user_pid,
                task_fields.user_tid,
                task_fields.user_ppid,
                task_fields.name,
            )
            yield 0, fields

    def run(self):
        decorate_comm = self.config.get("decorate_comm")

        headers = [
            ("OFFSET", format_hints.Hex),
            ("PID", int),
            ("TID", int),
            ("PPID", int),
            ("COMM", str),
        ]
        return renderers.TreeGrid(headers, self._generator(decorate_comm=decorate_comm))



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import Iterable, Generator, Tuple

from volatility3.framework import interfaces, symbols, exceptions
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist, vadinfo

vollog = logging.getLogger(__name__)


class Malfind(interfaces.plugins.PluginInterface):
    """Lists process memory ranges that potentially contain injected code."""

    _required_framework_version = (2, 22, 0)
    _version = (1, 1, 0)

    @classmethod
    def get_requirements(cls):
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract injected VADs",
                default=False,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="vadinfo", component=vadinfo.VadInfo, version=(2, 0, 0)
            ),
        ]

    @classmethod
    def is_vad_empty(cls, proc_layer, vad):
        """Check if a VAD region is either entirely unavailable due to paging,
        entirely consisting of zeros, or a combination of the two. This helps
        ignore false positives whose VAD flags match task._injection_filter
        requirements but there's no data and thus not worth reporting it.

        Args:
            proc_layer: the process layer
            vad: the MMVAD structure to test

        Returns:
            A boolean indicating whether a VAD is empty or not
        """

        CHUNK_SIZE = 0x1000
        all_zero_page = b"\x00" * CHUNK_SIZE

        offset = 0
        vad_length = vad.get_size()

        while offset < vad_length:
            next_addr = vad.get_start() + offset
            if (
                proc_layer.is_valid(next_addr, CHUNK_SIZE)
                and proc_layer.read(next_addr, CHUNK_SIZE) != all_zero_page
            ):
                return False
            offset += CHUNK_SIZE

        return True

    @classmethod
    def list_injections(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_layer_name: str,
        symbol_table: str,
        proc: interfaces.objects.ObjectInterface,
    ) -> Iterable[Tuple[interfaces.objects.ObjectInterface, bytes]]:
        for vad, data_object in cls.list_injection_sites(
            context, kernel_layer_name, symbol_table, proc
        ):
            yield (
                vad,
                data_object.context.layers[data_object.layer_name].read(
                    data_object.offset, data_object.length
                ),
            )

    @classmethod
    def list_injection_sites(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_layer_name: str,
        symbol_table: str,
        proc: interfaces.objects.ObjectInterface,
    ) -> Generator[
        Tuple[interfaces.objects.ObjectInterface, renderers.LayerData],
        None,
        None,
    ]:
        """Generate memory regions for a process that may contain injected
        code.

        Args:
            context: The context from which to retrieve required elements (layers, symbol tables)
            kernel_layer_name: The name of the kernel layer from which to read the VAD protections
            symbol_table: The name of the table containing the kernel symbols
            proc: an _EPROCESS instance

        Returns:
            An iterable of VAD instances and the first 64 bytes of data contained in that region
        """
        proc_id = "Unknown"
        try:
            proc_id = proc.UniqueProcessId
            proc_layer_name = proc.add_process_layer()
        except exceptions.InvalidAddressException as excp:
            vollog.debug(
                f"Process {proc_id}: invalid address {excp.invalid_address} in layer {excp.layer_name}"
            )
            return None

        proc_layer = context.layers[proc_layer_name]

        for vad in proc.get_vad_root().traverse():
            protection_string = vad.get_protection(
                vadinfo.VadInfo.protect_values(
                    context, kernel_layer_name, symbol_table
                ),
                vadinfo.winnt_protections,
            )
            write_exec = "EXECUTE" in protection_string and "WRITE" in protection_string
            dirty_page = None
            if not write_exec:
                """
                # Inspect "PAGE_EXECUTE_READ" VAD pages to detect
                # non-writable memory regions having been injected
                # using elevated WriteProcessMemory().
                """
                if "EXECUTE" in protection_string:
                    for page in range(
                        vad.get_start(), vad.get_end(), proc_layer.page_size
                    ):
                        try:
                            # If we have a dirty page in a non-writable "EXECUTE" region, it is suspicious.
                            if proc_layer.is_dirty(page):
                                dirty_page = page
                                break
                        except exceptions.InvalidAddressException:
                            # Abort as it is likely that other addresses in the same range will also fail.
                            break
                    if dirty_page is None:
                        continue
                else:
                    continue

            if (vad.get_private_memory() == 1 and vad.get_tag() == "VadS") or (
                vad.get_private_memory() == 0
                and protection_string != "PAGE_EXECUTE_WRITECOPY"
            ):
                if cls.is_vad_empty(proc_layer, vad):
                    continue

                if dirty_page is not None:
                    # Useful information to investigate the page content with volshell afterwards.
                    vollog.debug(
                        f"[proc_id {proc_id}] Found suspicious DIRTY + {protection_string} page at {hex(dirty_page)}",
                    )
                start = vad.get_start()
                length = 64
                data = renderers.LayerData(
                    context=context,
                    layer_name=proc_layer_name,
                    offset=start,
                    length=length,
                    no_surrounding=True,
                )
                yield (vad, data)

    def _generator(self, procs):
        # Determine if we're on a 32 or 64 bit kernel
        kernel = self.context.modules[self.config["kernel"]]

        # Set refined criteria to know when to add to "Notes" column
        refined_criteria = {
            b"MZ": "MZ header",
            b"\x55\x8b": "PE header",
            b"\x55\x48": "Function prologue",
            b"\x55\x89": "Function prologue",
        }

        is_32bit_arch = not symbols.symbol_table_is_64bit(
            context=self.context, symbol_table_name=kernel.symbol_table_name
        )

        for proc in procs:
            # By default, "Notes" column will be set to N/A
            process_name = utility.array_to_string(proc.ImageFileName)

            for vad, data_object in self.list_injection_sites(
                self.context, kernel.layer_name, kernel.symbol_table_name, proc
            ):
                notes = renderers.NotApplicableValue()
                # Check for unique headers and update "Notes" column if criteria is met
                data = data_object.context.layers[data_object.layer_name].read(
                    data_object.offset, data_object.length, True
                )
                if data[:2] in refined_criteria:
                    notes = refined_criteria[data[:2]]

                # If we're on a 64 bit kernel, we may still need 32 bit disasm due to wow64
                if is_32bit_arch or proc.get_is_wow64():
                    architecture = "intel"
                else:
                    architecture = "intel64"

                disasm = renderers.Disassembly(data, vad.get_start(), architecture)

                file_output = "Disabled"
                if self.config["dump"]:
                    file_output = "Error outputting to file"
                    try:
                        file_handle = vadinfo.VadInfo.vad_dump(
                            self.context, proc, vad, self.open
                        )
                        file_handle.close()
                        file_output = file_handle.preferred_filename
                    except (exceptions.InvalidAddressException, OverflowError) as excp:
                        vollog.debug(
                            f"Unable to dump PE with pid {proc.UniqueProcessId}.{vad.get_start():#x}: {excp}"
                        )

                yield (
                    0,
                    (
                        proc.UniqueProcessId,
                        process_name,
                        format_hints.Hex(vad.get_start()),
                        format_hints.Hex(vad.get_end()),
                        vad.get_tag(),
                        vad.get_protection(
                            vadinfo.VadInfo.protect_values(
                                self.context,
                                kernel.layer_name,
                                kernel.symbol_table_name,
                            ),
                            vadinfo.winnt_protections,
                        ),
                        vad.get_commit_charge(),
                        vad.get_private_memory(),
                        file_output,
                        notes,
                        data_object,
                        disasm,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Start VPN", format_hints.Hex),
                ("End VPN", format_hints.Hex),
                ("Tag", str),
                ("Protection", str),
                ("CommitCharge", int),
                ("PrivateMemory", int),
                ("File output", str),
                ("Notes", str),
                ("Hexdump", renderers.LayerData),
                ("Disasm", renderers.Disassembly),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=filter_func,
                )
            ),
        )
